import glob
import os
import cv2  # optional
#import cv2.cv as cv  # FIXME
import numpy
import urllib
import timeit
import time
import tempfile
#import multiprocessing
import signal 
import sys
from vipy.util import imresize, tempimage



class Camera(object):
    CAM = None
    FRAMERATE = False
    TIC = 0
    TOC = 0
    RESIZE = None
    GREY = None
    PROCESS = None
    
class Webcam(Camera):
    def __init__(self, framerate=False, resize=1, grey=False, url=0):
        self.CAM = cv2.VideoCapture(url)
        if not self.CAM.isOpened():
            self.CAM.open(url)
        self.FRAMERATE = framerate
        self.RESIZE = resize
        self.GREY = grey
        if framerate:
            self.TIC = timeit.default_timer()

    def __del__(self):
        self.CAM.release()
        return self
        
    def __iter__(self):
        return self

    def _read(self):
        #im = self.CAM.get()
        #return self.CAM.get() # HACK: for slow processing to get most recent image
        return self.CAM.read() # HACK: for slow processing to get most recent image

    def current(self):
        return self.next(grab=False)
                
    def next(self, grab=True):
        k = 0
        while not self.CAM.grab():
            k = k + 1
            #print '[webcam.camera][%d/%d]: invalid grab' % (k, 100)
            if k > 100:
                raise ValueError('Invalid Frame')                
                
        (rval, im) = self.CAM.retrieve()
#        if rval is False:
#            for i in range(0,60):
#                (rval, im) = self._retrieve()
#                if rval is True:
#                    break
        if rval is False:
            raise ValueError('Invalid Frame')
        if self.RESIZE != 1:
            im = imresize(im, self.RESIZE) 
        if self.GREY:
            im = cv2.cvtColor(im, cv2.COLOR_BGR2GRAY)
        if self.FRAMERATE:
            self.TOC = timeit.default_timer()
            print('[bobo.camera]: frame rate = ' + str(round(1.0/(self.TOC-self.TIC),1)) + ' Hz')
            self.TIC = self.TOC
        return im



class MotionStereo(Webcam):    
    IMCURRENT = None
    IMPREV = None
    
    def next(self):    
        if self.IMPREV is None:
            self.IMPREV = super(MotionStereo, self).next()
        else:
            self.IMPREV = self.IMCURRENT.copy()
        self.IMCURRENT = super(MotionStereo, self).next()
        return (self.IMCURRENT, self.IMPREV)

    
class Ipcam(Camera):
    TMPFILE = None
    def __init__(self, url, imfile=tempimage()):
        self.CAM = url
        self.TMPFILE = imfile
    
    def __iter__(self):
        return self

    def next(self):
        urllib.urlretrieve(self.CAM, self.TMPFILE)
        return cv2.imread(self.TMPFILE)  # numpy
  



#!/usr/bin/env python
"""FDDB dataset module"""

from __future__ import print_function
import os
import math
from collections import defaultdict
import bobo.app
from bobo.util import quietprint, remkdir, Stopwatch, saveas, filetail, filepath, filebase, islist
from bobo.image import Image, ImageDetection
from bobo.geometry import BoundingBox, Ellipse
from bobo.cache import Cache

from janus.detection import  DlibObjectDetector

class FDDB(object):
    """Manages the FDDB dataset: http://vis-www.cs.umass.edu/fddb"""

    def __init__(self, rootdir=None):
        self.root_dir = (rootdir if rootdir is not None
                         else os.path.join(bobo.app.datadir(), 'fddb'))
        self.img_dir = os.path.join(self.root_dir, 'img')
        self.folds_dir = os.path.join(self.root_dir, 'FDDB-folds')

    def cachecheck(self):
        if not os.path.isdir(os.path.join(self.img_dir, '2002')):
            """ The tarball contains multiple copies of some images with read-only permissions.
            This causes the tarfile module to throw an IOError partway through extraction when
            trying to overwite.  Ask the user to download and extract manually.  """

            data_url = 'http://tamaraberg.com/faceDataset/originalPics.tar.gz'
            raise ValueError('Download FDDB dataset manually and unpack to to "%s" ' % self.img_dir)

        if not os.path.isfile(os.path.join(self.folds_dir, 'FDDB-fold-01.txt')):
            remkdir(self.root_dir)
            folds_url = 'http://vis-www.cs.umass.edu/fddb/FDDB-folds.tgz'
            folds_sha1 = '94ce19ba3348425dfbb8bcbe55802490f8f152f8'
            folds_file = os.path.join(self.root_dir, filetail(folds_url))
            cache = Cache()
            cache.get(url=folds_url, sha1=folds_sha1, key=folds_file)
            cache.unpack(folds_file, unpackto=self.root_dir)
        return True


    def imset(self, fold=1):
        if not self.cachecheck():
            return
        fold_file = os.path.join(self.folds_dir, 'FDDB-fold-%02d.txt' % fold)
        imset = []
        with open(fold_file, 'r') as f:
            for line in f:
                filename = self.unkey(line.strip())
                imset.append(Image(filename=filename))
        return imset


    def groundtruth(self, fold=1):
        if not self.cachecheck():
            return
        fold_file = os.path.join(self.folds_dir,
                                 'FDDB-fold-%02d-ellipseList.txt' % fold)
        ground_truth = defaultdict(list)
        with open(fold_file, 'r') as f:
            for line in f:
                # image name
                filename = self.unkey(line.strip())
                dets = []
                # next line has truth count
                detcnt = next(f)
                for i in range(int(detcnt)):
                    # Detection
                    det = next(f)
                    (maj_rad, min_rad, angle, center_x, center_y,
                     score) = [float(z) for z in det.split()]
                    # FIXME: Assumes semi-major is vertical
                    xmin = center_x - min_rad
                    ymin = center_y - maj_rad
                    xmax = center_x + min_rad
                    ymax = center_y + maj_rad
                    ellipse = Ellipse(semi_major=maj_rad, semi_minor=min_rad, phi=angle,
                                    xcenter=center_x, ycenter=center_y)
                    bbox = ellipse.estimate_boundingbox()
                    bbox.score = 1.0
                    d = dict(ellipse=ellipse, bbox=bbox)
                    ground_truth[filename].append(d)
        return dict(ground_truth)


    def folds(self, folds=1):
        if not islist(folds):
            folds = [folds]
        truths = []
        imsets = []
        for fold in folds:
            truths.append(self.groundtruth(fold=fold))
            imsets.append(self.imset(fold=fold))
        return (imsets, truths)


    def detectionset(self, fold=1, name=None):
        imset = self.imagedetectionset(fold=fold, name=name)
        dets = defaultdict(list)

        for im in imset:
            fn = im.filename()
            dets[fn].append(im.boundingbox())
        return dets


    def imagedetectionset(self, fold=1, name=None):
        dets_dir = os.path.join(self.root_dir, name)
        fold_file = os.path.join(dets_dir, 'fold-%02d-out.txt' % fold)
        imset = []
        with open(fold_file, 'r') as f:
            for line in f: # We will advance this iterator inside the loop as well
                filename = self.unkey(line.strip())
                detcnt = next(f)
                for i in range(int(detcnt)):
                    det = next(f)
                    vals = det.split()
                    if len(vals) == 5:
                        (xmin, ymin, width, height, score) = [float(z) for z in vals]
                        xmax = xmin + width
                        ymax = ymin + height
                    else:
                        (smaj, smin, angle, center_x, center_y, score) = [float(z) for z in vals]
                        xmin = center_x - smin
                        ymin = center_y - smaj
                        xmax = center_x + smin
                        ymax = center_y + smaj
                    bbox = BoundingBox(xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, score=score)
                    imset.append(ImageDetection(filename=filename, bbox=bbox))
        return imset


    def key(self, filename):
        """FDDB_DIR/img/2002/08/11/big/img_591.jpg => 2002/08/11/big/img_591"""
        return '/'.join(filename.split('/')[-5:])[:-4]


    def unkey(self, key):
        """2002/08/11/big/img_591 => FDDB_DIR/img/2002/08/11/big/img_591.jpg"""
        imgname = key + '.jpg'
        filename = os.path.join(self.img_dir, imgname)
        return filename


    def writedetections(self, name, fold, detections,
                        write_images=False, do_ellipse=False):
        """ Takes a list of tuples where the first item is image key, and the
            second is a list of bounding box detections.  Writes output in FDDB
            detectionformat """

        outfile = os.path.join(self.root_dir, name, 'fold-%02d-out.txt' % fold)
        remkdir(bobo.util.filepath(outfile))
        with open(outfile, 'w') as f:
            for k, v in detections:
                key = self.key(k)
                print(key, file=f)
                print(str(len(v)), file=f)
                for d in v:
                    if do_ellipse:
                        """Bounding box format: semimajor_radius,
                        semiminor_radius, angle, center_x, center_y, score"""
                        cx, cy = d.centroid()
                        h = d.height()
                        w = d.width()
                        a = math.pi / 2 # semimajor is vertical
                        print('{0} {1} {2} {3} {4} {5}'.format(h/2, w/2, a, cx, cy, d.score), file=f)
                    else:
                        # Bounding box format: left, top, width, height, score
                        print('{0} {1} {2} {3} {4}'.format(d.xmin, d.ymin, d.width(), d.height(), d.score), file=f)
                if write_images:
                    image_dir = os.path.join(self.root_dir, name, filepath(key))
                    remkdir(image_dir)
                    im = Image(filename=k)
                    im.load()
                    for d in v:
                        im.drawbox(bbox=d)
                    fn = filebase(im.filename() + '.png')
                    out_img = os.path.join(image_dir, fn)
                    quietprint('Saving %s' % out_img, 2)
                    im.saveas(out_img)



"""
Copyright (c) 2010 Gary Wilson Jr. <gary.wilson@gmail.com> and contributers.

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

From:
http://pypi.python.org/pypi/python-archive/0.1
http://code.google.com/p/python-archive/

Changelog:
----------

* 2011/09/02: Cosmetic changes and add verbose kwarg to {Tar,Zip}Archive
classes by Nicolas Pinto <pinto@rowland.harvard.edu>

"""

import os
import tarfile
import zipfile
import sys

class ArchiveException(Exception):
    """Base exception class for all archive errors."""


class UnrecognizedArchiveFormat(ArchiveException):
    """Error raised when passed file is not a recognized archive format."""


def extract(archive_filename, output_dirname='./', verbose=True):
    """
    Unpack the tar or zip file at the specified `archive_filename` to the
    directory specified by `output_dirname`.
    """
    Archive(archive_filename).extract(output_dirname)


class Archive(object):
    """
    The external API class that encapsulates an archive implementation.
    """

    def __init__(self, file):
        self._archive = self._archive_cls(file)(file)

    @staticmethod
    def _archive_cls(file):
        cls = None
        if isinstance(file, str):
            filename = file
        else:
            try:
                filename = file.name
            except AttributeError:
                raise UnrecognizedArchiveFormat(
                    "File object not a recognized archive format.")
        base, tail_ext = os.path.splitext(filename.lower())
        cls = extension_map.get(tail_ext)
        if not cls:
            base, ext = os.path.splitext(base)
            cls = extension_map.get(ext)
        if not cls:
            raise UnrecognizedArchiveFormat(
                "Path not a recognized archive format: %s" % filename)
        return cls

    def extract(self, output_dirname=''):
        self._archive.extract(output_dirname)

    def list(self):
        self._archive.list()


class BaseArchive(object):
    """
    Base Archive class.  Implementations should inherit this class.
    """

    def extract(self):
        raise NotImplementedError

    def list(self):
        raise NotImplementedError


class ExtractInterface(object):
    """
    Interface class exposing common extract functionalities for
    standard-library-based Archive classes (e.g. based on modules like tarfile,
    zipfile).
    """

    def extract(self, output_dirname, verbose=True):
        if not verbose:
            self._archive.extractall(output_dirname)
        else:
            members = self.get_members()
            n_members = len(members)
            for mi, member in enumerate(members):
                self._archive.extract(member, path=output_dirname)
                extracted = mi + 1
                status = (r"Progress: %20i files extracted [%4.1f%%]"
                          % (extracted, extracted * 100. / n_members))
                status += chr(8) * (len(status) + 1)
                print(status, end=' ')
                sys.stdout.flush()
            print()


class TarArchive(ExtractInterface, BaseArchive):

    def __init__(self, filename):
        self._archive = tarfile.open(filename)

    def list(self, *args, **kwargs):
        self._archive.list(*args, **kwargs)

    def get_members(self):
        return self._archive.getmembers()

    
class ZipArchive(ExtractInterface, BaseArchive):

    def __init__(self, filename):
        self._archive = zipfile.ZipFile(filename)

    def list(self, *args, **kwargs):
        self._archive.printdir(*args, **kwargs)

    def get_members(self):
        return self._archive.namelist()


extension_map = {
    '.egg': ZipArchive,
    '.jar': ZipArchive,
    '.tar': TarArchive,
    '.tar.bz2': TarArchive,
    '.tar.gz': TarArchive,
    '.tgz': TarArchive,
    '.tz2': TarArchive,
    '.zip': ZipArchive,
}



import os
import csv
from bobo.cache import Cache
from bobo.util import remkdir, isstring
from bobo.viset.stream import ImageDetectionStream
import numpy as np

URL = 'http://www.vision.ee.ethz.ch/datasets_extra/ethz_shape_classes_v12.tgz'
SHA1 = 'ae9b8fad2d170e098e5126ea9181d0843505a84b'
SUBDIR = 'ETHZShapeClasses-V1.2'
LABELS = ['Applelogos','Bottles','Giraffes','Mugs','Swans']
VISET = 'ethzshapes'

cache = Cache()

def stream():
    csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if not os.path.isfile(csvfile):
        csvfile = export()        
    return ImageDetectionStream(csvfile)

def export():        
    # Fetch data necessary to initial construction
    pkgdir = cache.unpack(cache.get(URL, key=os.path.join(VISET, '%s.tgz' % VISET)), unpackto=os.path.join(cache.root(), VISET), sha1=SHA1, cleanup=True)
    categorydir = LABELS
    outfile = os.path.join(cache.root(), '%s.csv' % VISET)
                    
    # Write dataset
    with open(outfile, 'wb') as csvfile:
        f = csv.writer(csvfile, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)         
        for (idx_category, category) in enumerate(categorydir):
            imdir = os.path.join(pkgdir, SUBDIR, category)        
            for filename in os.listdir(imdir):
                if filename.endswith(".jpg") and not filename.startswith('.'):
                    # Write image
                    im = os.path.join(VISET, SUBDIR, category, filename)

                    # Write detections
                    gtfile = os.path.join(pkgdir, SUBDIR, category, os.path.splitext(os.path.basename(filename))[0] + '_' + category.lower()+'.groundtruth')
                    if not os.path.isfile(gtfile):
                        gtfile = os.path.join(pkgdir, SUBDIR, category, os.path.splitext(os.path.basename(filename))[0] + '_' + category.lower()+'s.groundtruth') # plural hack
                    for line in open(gtfile,'r'):
                        if line.strip() == '':
                            continue
                        (xmin,ymin,xmax,ymax) = line.strip().split()
                        f.writerow([im, category, xmin, ymin, xmax, ymax]);
        
    return outfile
    



import os
import numpy as np
from vipy.util import remkdir, imlist, filetail
import shutil
from vipy.util import istuple, islist, isnumpy, quietprint, remkdir, filebase, filetail
from vipy.image import Image
from vipy.video import VideoDetection
from vipy.show import savefig
from collections import defaultdict
import time
import PIL


def montage(imset, m, n, rows=None, cols=None, aspectratio=1, crop=False, skip=True, grayscale=True, do_plot=False, figure=None, border=0, border_bgr=(128,128,128), do_flush=False, verbose=False):
    """Montage image of images of size (m,n), such that montage has given aspect ratio.  Pass in iterable of imagedetection objects which is used to montage rowwise"""

    n_imgs = len(imset)
    M = int(np.ceil(np.sqrt(n_imgs)))
    N = M
    if aspectratio != 1:
        x = int(round((aspectratio*N-M)/(1+aspectratio)))
        N = N - x
        M = M + x
    elif rows is not None and cols is not None:
        N = rows
        M = cols
    padding = (M+1) * border
    size = (M * m + padding, N * n + padding)
    bc = border_bgr
    if grayscale:
        if islist(bc) or istuple(bc) or isnumpy(bc):
            bc = np.mean(bc)
        I = np.array(PIL.Image.new(mode='L', size=size, color=bc))
    else:
        I = np.array(PIL.Image.new(mode='RGB', size=size, color=bc))
    k = 0
    for j in range(N):
        for i in range(M):
            if k >= n_imgs: 
                break
            sliceM, sliceN = i * (m+border) + border, j * (n+border) + border
            try:
                if crop:
                    if not imset[k].bbox.valid():
                        print('[janus.visualize.montage] invalid bounding box "%s" ' % str(imset[k].bbox))
                        if skip == False:
                            print('[janus.visualize.montage] using original image')
                            if grayscale:
                                im = imset[k].grayscale().resize(n,m)._array                                                
                            else:
                                im = imset[k].resize(n,m)._array
                        else:
                            raise
                    else:
                        if grayscale:
                            im = imset[k].grayscale().crop(imset[k].bbox).resize(n,m)._array
                        else:
                            im = imset[k].crop(imset[k].bbox).resize(n,m)._array
                else:
                    if grayscale:
                        im = imset[k].grayscale().resize(n,m)._array  # m=width, n=height
                    else:
                        im = imset[k].resize(n,m)._array
       
                if im.dtype == np.float32:
                    if im.max() <= 1.0:
                        im *= 255.0
                    im = im.astype(np.uint8)

                I[sliceN:sliceN + n, sliceM:sliceM + m] = im
                
            except KeyboardInterrupt:
                raise
            except:
                print('[janus.visualize.montage] skipping...')
                if skip:
                    pass
                else:
                    raise

            if do_flush:
                imset[k].flush()  # clear memory
            if verbose and ((k % 100) == 0):
                print('[janus.visualize.montage][%d/%d] processing...' % (k, n_imgs))
            
            k += 1

    if k == 0:
        print('[janus.visualize.montage] Warning: No images were processed')

    if do_plot is True:
        im = Image('')
        im._array = I
        # HACK: float(0-255) graycale images display incorrectly
        if grayscale:
            im.preprocess().rgb().show(figure=figure)
        else:
            im.show(figure=figure)

    return I       


def imagelist(list_of_image_files, outdir, title='Image Visualization', imagewidth=64):
    """Given a list of image filenames wth absolute paths, copy to outdir, and create an index.html file that visualizes each"""
    k_divid = 0;
    
    # Create summary page to show precomputed images
    outdir = remkdir(outdir)
    filename = os.path.join(remkdir(outdir), 'index.html');
    f = open(filename,'w')
    f.write('<!DOCTYPE html>\n')
    f.write('<html>\n')
    f.write('<body>\n')
    f.write('<div id="container" style="width:2400px">\n')
    f.write('<div id="header">\n')
    f.write('<h1 style="margin-bottom:0;">Title: %s</h1><br>\n' % title)
    localtime = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(time.time()))
    f.write('Summary HTML generated on %s<br>\n' % localtime)
    f.write('Number of Images: %d<br>\n' % len(list_of_image_files))
    f.write('</div>\n')
    f.write('<br>\n')
    f.write('<hr>\n')
    f.write('<div id="%04d" style="float:left;">\n' % k_divid);  k_divid = k_divid + 1;
    
    # Generate images and html
    for (k, imsrc) in enumerate(list_of_image_files):
        shutil.copyfile(imsrc, os.path.join(outdir, filetail(imsrc)))
        imdst = filetail(imsrc)
        f.write('<p>\n</p>\n')        
        f.write('<b>Filename: %s</b><br>\n' % imdst) 
        f.write('<br>\n')                
        f.write('<img src="%s" alt="image" width=%d/>\n' % (imdst, imagewidth))
        f.write('<p>\n</p>\n')
        f.write('<hr>\n')
        f.write('<p>\n</p>\n')        
                
    f.write('</div>\n')
    f.write('</body>\n')
    f.write('</html>\n')    
    f.close()
    return filename


def imagetuplelist(list_of_tuples_of_image_files, outdir, title='Image Visualization', imagewidth=64):
    """Imageset but put tuples on same row"""
    k_divid = 0;
    
    # Create summary page to show precomputed images
    outdir = remkdir(outdir)
    filename = os.path.join(remkdir(outdir), 'index.html');
    f = open(filename,'w')
    f.write('<!DOCTYPE html>\n')
    f.write('<html>\n')
    f.write('<body>\n')
    f.write('<div id="container" style="width:2400px">\n')
    f.write('<div id="header">\n')
    f.write('<h1 style="margin-bottom:0;">Title: %s</h1><br>\n' % title)
    localtime = time.strftime("%Y-%m-%d %H:%M:%S",time.localtime(time.time()))
    f.write('Summary HTML generated on %s<br>\n' % localtime)
    f.write('Number of Tuples: %d<br>\n' % len(list_of_tuples_of_image_files))
    f.write('</div>\n')
    f.write('<br>\n')
    f.write('<hr>\n')
    f.write('<div id="%04d" style="float:left;">\n' % k_divid);  k_divid = k_divid + 1;
    
    # Generate images and html
    for (k, imsrclist) in enumerate(list_of_tuples_of_image_files):
        f.write('<p>\n</p>\n') 
        for imsrc in imsrclist:       
            shutil.copyfile(imsrc, os.path.join(outdir, filetail(imsrc)))
            imdst = filetail(imsrc)
            f.write('<b>Filename: %s</b><br>\n' % imdst) 
        f.write('<p>\n</p>\n')     
        f.write('<br>\n')                   
        for imsrc in imsrclist:
            imdst = filetail(imsrc)
            f.write('<img src="%s" alt="image" width=%d/>' % (imdst, imagewidth))
        f.write('\n<p>\n</p>\n')
        f.write('<hr>\n')
        f.write('<p>\n</p>\n')        
                
    f.write('</div>\n')
    f.write('</body>\n')
    f.write('</html>\n')    
    f.close()
    return filename



import os
from bobo.util import remkdir, quietprint, filetail
from bobo.image import ImageCategory
import bobo.app

DEV_IMAGES_URL = 'http://www.cs.columbia.edu/CAVE/databases/pubfig/download/dev_urls.txt'
#DEV_PEOPLE_URL = 'http://www.cs.columbia.edu/CAVE/databases/pubfig/download/dev_people.txt'
#EVAL_PEOPLE_URL = 'http://www.cs.columbia.edu/CAVE/databases/pubfig/download/eval_people.txt'
EVAL_IMAGES_URL = 'http://www.cs.columbia.edu/CAVE/databases/pubfig/download/eval_urls.txt'

DEV_IMAGES_SHA1 = '9eb10c01d46c5d06a8f70b9b8a9ff6b8fe4b0e41';
EVAL_IMAGES_SHA1 = '0fd4cfc464993909c45f9bce1322747c9a9baef9';


class PubFig(object):
    def __init__(self, datadir=None):
        self.datadir = bobo.app.datadir() if datadir is None else datadir
        self.imdir = os.path.join(self.datadir, 'pubfig')
        
        if not os.path.isfile(os.path.join(self.imdir, 'dev_urls.txt')):
            raise ValueError('Download PubFig dev_urls.txt manually and save to "%s" ' % self.imdir)
        
        
    def __repr__(self):
        return str('<viset.pubfig: %s>' % self.imdir)

    
    def rdd(self, sparkContext=None, ignore=True, fetch=True):
        """Create a resilient distributed dataset"""
        sparkContext = bobo.app.init('viset_pubfig') if sparkContext is None else sparkContext
        csvfile = os.path.join(self.imdir, 'dev_urls.txt')
        return (sparkContext.textFile(csvfile)  # RDD of textfile lines
                            .map(lambda row: row.decode('utf-8').rstrip().split())  # CSV to list
                            .filter(lambda r: r[0] not in ['#', '\u#'])
                            .map(lambda x: ImageCategory(url=x[3], filename=os.path.join(self.imdir, filetail(x[3])),  category='%s_%s' % (x[0], x[1]))))
        
    def devlist(self):
        """Return interface objects for this dataset"""
        return self.rdd().collect()

    def download(self):
        return self.rdd().map(lambda im: im.download(ignoreErrors=True, timeout=10)).collect()




import csv

class ImageStream(object):
    def __init__(self, csvfile, parser, delimiter=' ', quotechar='|', rowstart=1):        
        self._csvfile = csvfile
        self._parser = parser
        self._delimiter = delimiter
        self._quotechar = quotechar
        self._rowstart = rowstart
            
    def __iter__(self):
        self._f_csv = open(self._csvfile, 'rb')
        self._reader = csv.reader(self._f_csv, delimiter=self._delimiter, quotechar=self._quotechar)  # reopen
        if self._rowstart > 1:
            for k in range(1, self._rowstart):
                self._reader.next()
        return self

    def __getitem__(self, item):
        # Inefficient method for random stream access
        f = open(self._csvfile, 'r')
        for i in range(max(item,1)):
            line = f.readline()
            if len(line) == 0:
                raise IndexError('Invalid index "%d"' % item)  # end of file
        f.close()
        return self._parser(line)

    def next(self):
        row = self._reader.next()
        return self._parser(row)
    

class VideoStream(ImageStream):
    _f_csv = None
    firstframe = None
    
    def __init__(self, csvfile, keyby, parser, delimiter=' ', quotechar='|', rowstart=1):        
        self._csvfile = csvfile
        self._parser = parser
        self._delimiter = delimiter
        self._quotechar = quotechar
        self._rowstart = rowstart
        self._keyby = keyby

    def __iter__(self):
        self._f_csv = open(self._csvfile, 'rb')
        self._reader = csv.reader(self._f_csv, delimiter=self._delimiter, quotechar=self._quotechar)  # reopen
        if self._rowstart > 1:
            for k in range(1, self._rowstart):
                self._reader.next()
        return self
        
    def __getitem__(self, item):
        # Inefficient method for random stream access
        self._rowstart = item
        return self._parser( [self.__iter__().next()] )   # return singleton list for consistency with next()
        
    def next(self):
        """Read at least one frame in row order until keyby field changes, then parse the list of rows into a video object"""
        is_finalframe = False
        firstframe = self.firstframe if self.firstframe is not None else self._reader.next()
        firstkey = self._keyby(firstframe)
        frames = [firstframe]
        while not is_finalframe:
            row = self._reader.next()
            framekey = self._keyby(row)
            if framekey == firstkey:
                frames.append(row)
            else:
                is_finalframe = True
            self.firstframe = row            
        return self._parser(frames)
    



import os
from glob import glob
from bobo.util import remkdir, isstring, quietprint, islist, tolist, loadmat
from bobo.image import ImageDetection, ImageCategory
from bobo.video import VideoDetection
import bobo.app


    
class YouTubeFaces(object):
    def __init__(self, datadir=None, parsed=None):
        self.datadir = bobo.app.datadir() if datadir is None else datadir
        self.ytfdir = os.path.join(self.datadir, 'YouTubeFaces')
        
        if not os.path.isdir(os.path.join(self.ytfdir, 'frame_images_DB')):
            raise ValueError('Download YouTubeFaces dataset manually and unpack to to "%s", pass in datadir, or set $JANUS_DATA ' % self.ytfdir)        

        self._parsed = parsed  # user saved output from self.parse for optional caching

        
    def __repr__(self):
        return str('<viset.youtubefaces: %s>' % self.ytfdir)
    

    def parse(self, max_num_subjects=None):
        """Parse youtubefaces into a list of ImageDetections"""
        
        if self._parsed is not None:
            return self._parsed
                
        # Write images and annotations
        # The data in this file is in the following format:
        # filename,[ignore],x,y,width,height,[ignore],[ignore]
        # where:
        # x,y are the center of the face and the width and height are of the rectangle that the face is in.
        # For example:
        # $ head -3 Richard_Gere.labeled_faces.txt
        # Richard_Gere\3\3.618.jpg,0,262,165,132,132,0.0,1
        # Richard_Gere\3\3.619.jpg,0,260,164,131,131,0.0,1
        # Richard_Gere\3\3.620.jpg,0,261,165,129,129,0.0,1

        imlist = []
        categorydir = os.path.join(self.ytfdir, 'frame_images_DB')
        for i,infilename in enumerate(glob('%s/*labeled_faces.txt'%categorydir)):
            if max_num_subjects and max_num_subjects == i: break
            print '[viset.youtubefaces:] parsing "%s" ' % infilename
            with open(infilename, 'r') as infile:
                for line in infile:
                    ( imname, ignore, x_ctr, y_ctr, w, h, ignore, ignore) = line.split(',')
                    imname = imname.replace('\\', '/')
                    category = imname.split('/')[0]
                    xmin = int(x_ctr) - int(float(w) / 2.0)
                    ymin = int(y_ctr) - int(float(h) / 2.0)
                    xmax = int(xmin) + int(w)
                    ymax = int(ymin) + int(h)
                    p = os.path.join(categorydir, imname) 
                
                    (category,videoid,filename) = imname.split('/')
                    mediaID = '%s_%s' % (category, videoid)
                    imlist.append(ImageDetection(filename=p, category=category, xmin=int(xmin), ymin=int(ymin), xmax=int(xmax), ymax=int(ymax)).setattribute('MEDIA_ID', str(mediaID)));

        # Done!
        self._parsed = imlist
        return imlist
    
    def subjects(self, sparkContext=None, max_num_subjects=None):
        return self.rdd(sparkContext=sparkContext, minPartitions=16, max_num_subjects=max_num_subjects).map(lambda v: v.category()).distinct().collect()
    
    def images(self, max_num_subjects=None):
        return self.parse(max_num_subjects=max_num_subjects)
            
    def videos(self, sparkContext=None, max_num_subjects=None):
        return self.rdd(sparkContext=sparkContext, max_num_subjects=max_num_subjects).collect()
                
    def rdd(self, sparkContext=None, appname='viset_youtubefaces', minPartitions=None, max_num_subjects=None):
        """Create a resilient distributed dataset"""
        sparkContext = bobo.app.init(appname) if sparkContext is None else sparkContext                                            
        return (sparkContext.parallelize(self.parse(max_num_subjects=max_num_subjects), minPartitions)
                            .map(lambda im: (im.attributes['MEDIA_ID'], im))  # keyby media ID
                            .reduceByKey(lambda a,b: tolist(a)+tolist(b))  # group all frames of video together by videoID key into giant concatenated list of frames
                            .map(lambda (k,fr): VideoDetection(frames = fr, attributes=fr[0].attributes, category=fr[0].category())))  # construct video object from grouped frames

    def splits(self):
        mat = loadmat(os.path.join(self.ytfdir, 'meta_data', 'meta_and_splits.mat'))
        return mat['Splits']  # splits are indexed into mat['mat_names'] dictionary, parse me!

        
class YouTubeFacesAligned(YouTubeFaces):
    def parse(self, max_num_subjects=None):
        """Parse youtubefaces into a list of ImageDetections"""
        
        if self._parsed is not None:
            return self._parsed
                
        imlist = []
        categorydir = os.path.join(self.ytfdir, 'frame_images_DB')
        for i,nfilename in enumerate(glob('%s/*labeled_faces.txt'%categorydir)):
            if max_num_subjects and max_num_subjects == i: break
            print '[viset.youtubefaces:] parsing "%s" ' % infilename
            with open(infilename, 'r') as infile:
                for line in infile:
                    ( imname, ignore, x_ctr, y_ctr, w, h, ignore, ignore) = line.split(',')
                    imname = imname.replace('\\', '/')
                    category = imname.split('/')[0]
                    (category,videoid,filename) = imname.split('/')                    
                    p = os.path.join(self.ytfdir, 'aligned_images_DB', category, videoid, 'aligned_detect_%s' % filename)                
                    mediaID = '%s_%s' % (category, videoid)
                    imlist.append(ImageDetection(filename=p, category=category).setattribute('MEDIA_ID', str(mediaID)))  # FIXME: bounding boxes?

        # Done!
        self._parsed = imlist
        return imlist
    



import os
import numpy as np
from bobo.util import remkdir
import gzip, struct
from array import array
from bobo.image import Image, ImageCategory
from itertools import product, combinations
from bobo.app import setverbosity

TRAIN_IMG_URL = 'http://yann.lecun.com/exdb/mnist/train-images-idx3-ubyte.gz'
TRAIN_IMG_SHA1 =  '6c95f4b05d2bf285e1bfb0e7960c31bd3b3f8a7d'
TRAIN_LBL_URL = 'http://yann.lecun.com/exdb/mnist/train-labels-idx1-ubyte.gz'
TRAIN_LBL_SHA1 = '2a80914081dc54586dbdf242f9805a6b8d2a15fc'
TEST_IMG_URL = 'http://yann.lecun.com/exdb/mnist/t10k-images-idx3-ubyte.gz'
TEST_IMG_SHA1 = 'c3a25af1f52dad7f726cce8cacb138654b760d48'
TEST_LBL_URL = 'http://yann.lecun.com/exdb/mnist/t10k-labels-idx1-ubyte.gz'
TEST_LBL_SHA1 = '763e7fa3757d93b0cdec073cef058b2004252c17'

class MNIST(object):
    def __init__(self, outdir):
        """download URLS above to outdir, then run export()"""        
        self.outdir = remkdir(outdir)
        if not self._downloaded():
            print '[viset.mnist]: downloading MNIST to "%s"' % self.outdir
            setverbosity(1)                        
            self._wget()
            print '[viset.mnist]: exporting MNIST to "%s"' % self.outdir            
            self._export()

    def _downloaded(self):
        gzip_downloaded = (os.path.exists(os.path.join(self.outdir, 'train-images-idx3-ubyte.gz')) and
                os.path.exists(os.path.join(self.outdir, 'train-labels-idx1-ubyte.gz')) and
                os.path.exists(os.path.join(self.outdir, 't10k-images-idx3-ubyte.gz')) and
                os.path.exists(os.path.join(self.outdir, 't10k-labels-idx1-ubyte.gz')))             
        unpacked_downloaded = (os.path.exists(os.path.join(self.outdir, 'train-images-idx3-ubyte')) and
                os.path.exists(os.path.join(self.outdir, 'train-labels-idx1-ubyte')) and
                os.path.exists(os.path.join(self.outdir, 't10k-images-idx3-ubyte')) and
                os.path.exists(os.path.join(self.outdir, 't10k-labels-idx1-ubyte')))             
        return (unpacked_downloaded or gzip_downloaded)

            
    def _wget(self):
        os.system('wget --directory-prefix=%s %s' % (self.outdir, TRAIN_IMG_URL))
        os.system('wget --directory-prefix=%s %s' % (self.outdir, TRAIN_LBL_URL))
        os.system('wget --directory-prefix=%s %s' % (self.outdir, TEST_IMG_URL))
        os.system('wget --directory-prefix=%s %s' % (self.outdir, TEST_LBL_URL))
    
    def _labels(self, gzfile):
        with gzip.open(gzfile, 'rb') as file:
            magic, size = struct.unpack(">II", file.read(8))
            if magic != 2049:
                raise ValueError('Magic number mismatch, expected 2049,'
                                'got %d' % magic)
            labels = array("B", file.read())
        return labels

    def _imread(self, dataset, index):
        """Read MNIST encoded images, adapted from: https://github.com/sorki/python-mnist/blob/master/mnist/loader.py"""
        gzfile = None
        
        with gzip.open(gzfile, 'rb') as file:
            magic, size, rows, cols = struct.unpack(">IIII", file.read(16))
            if magic != 2051:
                raise ValueError('Magic number mismatch, expected 2051, got %d' % magic)
            file.seek(index*rows*cols + 16)
            image = numpy.asarray(array("B", file.read(rows*cols)).tolist())
            return numpy.reshape(image, (rows,cols))
		
    def trainset(self):
        y_train = self._labels(os.path.join(self.outdir, 'train-labels-idx1-ubyte.gz')).tolist() 
        return [ImageCategory(filename=os.path.join(self.outdir, 'train', '%06d.png' % k), category=str(y)) for (k, y) in enumerate(y_train) if k <= 0.9*len(y_train)]

    def subset(self, dataset, N=256):
        """Balanced subset"""
        D = {int(c):[im for im in dataset if im.category()==str(c)] for c in self.categories()}
        V = []
        for c in self.categories():
            V = V + (np.random.choice(D[c], np.ceil(N/10.0))).tolist()
        np.random.shuffle(V)
        return V[0:N]
        
    def validationset(self, N=None):
        y_val = self._labels(os.path.join(self.outdir, 'train-labels-idx1-ubyte.gz')).tolist() 
        return [ImageCategory(filename=os.path.join(self.outdir, 'train', '%06d.png' % k), category=str(y)) for (k, y) in enumerate(y_val) if k > 0.9*len(y_val)]
                                            
    def categories(self):
        return range(0,10)
    
    def pairs(self, dataset, N=1E6):
        """Balanced set of randomly selected N positive and negative pairs given dataset"""
        D = {int(c):[im for im in dataset if im.category()==str(c)] for c in self.categories()}
        M = np.maximum(N, 256)
        T = []
        for (p,q) in combinations(self.categories(), 2):  # 45 category combinations 
            T = T + zip(*(np.random.choice(D[q], size=np.ceil(M/(45*4))), np.random.choice(D[p], size=np.ceil(M/(45*4)))))  # different (p,q)
            T = T + zip(*(np.random.choice(D[p], size=np.ceil(M/(45*4))), np.random.choice(D[q], size=np.ceil(M/(45*4)))))  # different (q,p)
            T = T + zip(*(np.random.choice(D[p], size=np.ceil(M/(45*4))), np.random.choice(D[p], size=np.ceil(M/(45*4)))))  # same (p,p)
            T = T + zip(*(np.random.choice(D[q], size=np.ceil(M/(45*4))), np.random.choice(D[q], size=np.ceil(M/(45*4)))))  # same (q,q)
        return T[0:int(N)]

    def triplets(self, dataset, N=1E6):
        """Balanced set of randomly selected N positive and negative pairs given dataset"""
        D = {int(c):[im for im in dataset if im.category()==str(c)] for c in self.categories()}
        M = np.maximum(N, 256)
        T = []
        for (p,q) in combinations(self.categories(), 2):  # 45 category combinations 
            T = T + zip(*(np.random.choice(D[p], size=np.ceil(M/(45*2))), np.random.choice(D[p], size=np.ceil(M/(45*2))), np.random.choice(D[q], size=np.ceil(M/(45*2)))))  # (p,p,q)
            T = T + zip(*(np.random.choice(D[q], size=np.ceil(M/(45*2))), np.random.choice(D[q], size=np.ceil(M/(45*2))), np.random.choice(D[p], size=np.ceil(M/(45*2)))))  # different (q,q,p)
        return T[0:int(N)]
        
    def testset(self):
        y_test = self._labels(os.path.join(self.outdir, 't10k-labels-idx1-ubyte.gz')).tolist() 
        return [ImageCategory(filename=os.path.join(self.outdir, 'test', '%06d.png' % k), category=str(y)) for (k, y) in enumerate(y_test)]

    def dataset(self):
        return (self.trainset(), self.validationset(), self.testset())
    
    def _export(self):
        # Fetch data necessary to initial construction (do not gunzip!)
        test_img_file = os.path.join(self.outdir, 't10k-images-idx3-ubyte.gz')
        train_img_file = os.path.join(self.outdir, 'train-images-idx3-ubyte.gz')
        
        # Write dataset as 28x28 PNG
        remkdir(os.path.join(self.outdir, 'train'))
        remkdir(os.path.join(self.outdir, 'test'))        
        with gzip.open(train_img_file, 'rb') as gzfile:
            magic, size, rows, cols = struct.unpack(">IIII", gzfile.read(16))
            if magic != 2051:
                raise ValueError('Magic number mismatch, expected 2051, got %d' % magic)

            for k in range(60000):
                img = np.asarray(array("B", gzfile.read(rows*cols)).tolist()).reshape( (rows, cols) ).astype(np.uint8)
                im = Image().array(img).resize(rows=32).saveas(os.path.join(self.outdir, 'train', '%06d.png' % k))
                if (k % 10000) == 0:
                    print '[viset.mnist][%d/%d]: exporting training images to "%s"' % (k, 60000, os.path.join(self.outdir, 'train'))
    
        with gzip.open(test_img_file, 'rb') as gzfile:
            magic, size, rows, cols = struct.unpack(">IIII", gzfile.read(16))
            if magic != 2051:
                raise ValueError('Magic number mismatch, expected 2051, got %d' % magic)
                
            for k in range(10000):
                img = np.asarray(array("B", gzfile.read(rows*cols)).tolist()).reshape( (rows, cols) ).astype(np.uint8)
                im = Image().array(img).resize(rows=32).saveas(os.path.join(self.outdir, 'test', '%06d.png' % k))
                if (k % 10000) == 0:            
                    print '[viset.mnist][%d/%d]: exporting testing images to "%s"' % (k, 10000, os.path.join(self.outdir, 'test'))
        
        # Cleanup    
        return self
    
    



import argparse
import base64
import csv
import os
from bobo.util import remkdir, dirlist, imlist, writecsv, readcsv

def extract(tsvfile, outdir):
    """https://github.com/cmusatyalab/openface/blob/master/data/ms-celeb-1m/extract.py"""
    with open(tsvfile, 'r') as tsvF:
        reader = csv.reader(tsvF, delimiter='\t')
        i = 0
        for row in reader:
            MID, imgSearchRank, faceID, data = row[0], row[1], row[4], base64.b64decode(row[-1])

            saveDir = os.path.join(outdir, MID)
            savePath = os.path.join(saveDir, "{}-{}.jpg".format(imgSearchRank, faceID))
            
            remkdir(saveDir)
            with open(savePath, 'wb') as f:
                f.write(data)
                
            i += 1
                
            if i % 1000 == 0:
                print("Extracted {} images.".format(i))


def export(tsvfile, tsvnames, outdir, csvfile):

    csvlist = []
    d_mid_to_name = {x[0]:x[1] for x in readcsv(tsvnames, separator='\t')}
    with open(tsvfile, 'r') as tsvF:
        reader = csv.reader(tsvF, delimiter='\t')
        i = 0
        for row in reader:
            MID, imgSearchRank, faceID, data = row[0], row[1], row[4], base64.b64decode(row[-1])

            saveDir = os.path.join(outdir, MID)
            savePath = os.path.join(saveDir, "{}-{}.jpg".format(imgSearchRank, faceID))
            
            i += 1
                
            csvlist.append( (savePath, d_mid_to_name[MID]) )
            if i % 100 == 0:
                print "[msceleb.csv][%d]: Extracting CSV (%s,%s,%s)" % (i,savePath,MID,d_mid_to_name[MID])

    print writecsv(csvlist, csvfile)



import os
import PIL
from vipy.show import imshow, imbbox, savefig
from vipy.util import isnumpy, quietprint, isurl, isimageurl, islist, \
    fileext, tempimage, mat2gray, imwrite, imwritejet, imwritegray, tmpjpg, tempjpg, imresize, imrescale, bgr2gray, gray2rgb, bgr2rgb, rgb2bgr, gray2hsv, bgr2hsv
from vipy.geometry import BoundingBox, similarity_imtransform, \
    similarity_imtransform2D, imtransform, imtransform2D
import vipy.dataset.download
import urllib.request
import urllib.error
import urllib.parse
import http.client as httplib
#import cv2
import copy
import numpy as np
import shutil
import io



# FIX <urlopen error [SSL: CERTIFICATE_VERIFY_FAILED] certificate
# verify failed (_ssl.c:581)>
# http://stackoverflow.com/questions/27835619/ssl-certificate-verify-failed-error
import ssl
if hasattr(ssl, '_create_unverified_context'):
    ssl._create_default_https_context = ssl._create_unverified_context


class Image(object):
    """Standard class for images"""
    def __init__(self, filename=None, url=None, array=None, attributes=None):
        # Private attributes
        self._ignoreErrors = False  # ignore errors during fetch (broken links)
        self._urluser = None      # basic authentication set with url() method
        self._urlpassword = None  # basic authentication set with url() method
        self._urlsha1 = None      # file hash if known
        self._filename = None   # Local filename
        self._url = None        # URL to download
        self._loader = None     # lambda function to load an image, set with loader() method

        # Initialization
        self._filename = filename
        if url is not None:
            assert isurl(url), 'Invalid URL'
        self._url = url
        if array is not None:
            assert isnumpy(array), 'Invalid Array'
        self._array = array   

        # Public attributes: passed in as a dictionary
        self.attributes = attributes  

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        """Yield single image for consistency with videos and templates"""
        yield self

    def __len__(self):
        """Images have length 1 always"""
        return 1

    def __repr__(self):
        if self.isloaded():
            return str('<vipy.image: height=%d, width=%d, color=%s>' % (
                self._array.shape[0], self._array.shape[1],
                str(self.getattribute('colorspace'))))
        elif self._filename is not None and self._url is None:
            return str("<vipy.image: filename='%s'>" % str(self._filename))
        elif self._url is not None and self._filename is None:
            return str("<vipy.image: url='%s'>" % str(self._url))
        elif self._url is not None and self._filename is not None:
            return str("<vipy.image: url='%s', filename='%s'>" % (str(self._url), str(self._filename)))
        else:            
            raise ValueError('vipy.image.Image must be constructed with an input')  # should never get here

    def loader(self, f):
        """Lambda function to load an image filename to a numpy array"""
        self._loader = f
        return self


    def load(self, ignoreErrors=False, asRaw=False):
        """Load image to cached private '_array' attribute and return numpy array"""
        try:
            # Return previously loaded image
            if self._array is not None:
                return self._array

            # Download URL to filename
            if self._url is not None:
                if self._filename is None and isimageurl(self._url):
                    self._filename = tempimage(fileext(self._url))
                elif self._filename is None:
                    self._filename = tempjpg()  # guess .jpg
                self.download(ignoreErrors=ignoreErrors)

            # Load filename to numpy array
            if fileext(self._filename) == '.npz':
                if self._loader is None:
                    raise ValueError('Must define a customer loader for '
                                     '.npz file format')
                self._array = self._loader(self._filename)
            else:
                # BGR color order!
                self._array = np.array(PIL.Image.open(self._filename))[:,:,::-1]  
                #self._array = cv2.imread(self._filename,
                #                       cv2.CV_LOAD_IMAGE_UNCHANGED) \
                #                       if asRaw else cv2.imread(self._filename)

            if self._array is None:
                if fileext(self._filename) == '.gif':
                    #quietprint('[vipy.image][WARNING]: IO error - could '
                    #           'not load "%s" using opencv, '
                    #           'falling back on PIL ' %
                    #           self._filename, 1)
                    # Convert .gif to luminance (grayscale) and export
                    # as numpy array
                    self._array = np.array(
                        PIL.Image.open(self._filename).convert('L'))
                else:
                    raise ValueError('invalid image file "%s"' % self._filename)

            # Image Atributes
            self.setattribute('colorspace',
                              'bgr' if self._array.ndim == 3 else 'gray')

        except IOError:
            if self._ignoreErrors or ignoreErrors:
                quietprint('[vipy.image][WARNING]: IO error - '
                           'Invalid image file, url or invalid write '
                           'permissions "%s" ' %
                           self.filename(), True)
                self._array = None
            else:
                raise

        except KeyboardInterrupt:
            raise

        except Exception:
            if self._ignoreErrors or ignoreErrors:
                quietprint('[vipy.image][WARNING]: '
                           'load error for image "%s"' %
                           self.filename(), verbosity=2)
                self._array = None
            else:
                raise

        return self._array

    def download(self, ignoreErrors=False, timeout=10):
        """Download URL to filename provided by constructor, or to temp filename"""
        if self._url is None or not isurl(str(self._url)):
            raise ValueError('[vipy.image.download][ERROR]: '
                             'Invalid URL "%s" ' % self._url)
        if self._filename is None:
            self._filename = tempimage(fileext(self._url))
        try:
            url_scheme = urllib.parse.urlparse(self._url)[0]
            if url_scheme in ['http', 'https']:
                #quietprint('[vipy.image.download]: '
                #           'downloading "%s" to "%s" ' %
                #           (self._url, self._filename), verbosity=1)
                vipy.dataset.download.download(self._url,
                                        self._filename,
                                        verbose=False,
                                        timeout=timeout,
                                        sha1=self._urlsha1,
                                        username=self._urluser,
                                        password=self._urlpassword)
            elif url_scheme == 'file':
                #quietprint('[vipy.image.download]: copying "%s" to "%s" ' %
                #           (self._url, self._filename), verbosity=2)
                shutil.copyfile(self._url, self._filename)
            elif url_scheme == 'hdfs':
                raise NotImplementedError('FIXME: support for '
                                          'hadoop distributed file system')
            else:
                raise NotImplementedError(
                    'Invalid URL scheme "%s" for URL "%s"' %
                    (url_scheme, self._url))

        except (httplib.BadStatusLine,
                urllib.error.URLError,
                urllib.error.HTTPError):
            if self._ignoreErrors or ignoreErrors:
                quietprint('[vipy.image][WARNING]: download failed - '
                           'ignoring image', 1)
                self._array = None
            else:
                raise

        except IOError:
            if self._ignoreErrors or ignoreErrors:
                quietprint('[vipy.image][WARNING]: IO error - '
                           'Invalid image file, url or '
                           'invalid write permissions "%s" ' %
                           self.filename(), True)
                self._array = None
            else:
                raise

        except KeyboardInterrupt:
            raise

        except Exception:
            if self._ignoreErrors or ignoreErrors:
                quietprint('[vipy.image][WARNING]: '
                           'load error for image "%s"' %
                           self.filename(), verbosity=2)
            else:
                raise
        self.flush()
        return self

    def flush(self):
        """Remove cached numpy array"""
        self._array = None
        return self

    def isloaded(self):
        return self._array is not None

    def show(self, colormap=None, figure=None, flip=True):
        assert self.isloaded(), 'Image not loaded'
        #quietprint('[vipy.image][%s]: displaying image' %
        #           (self.__repr__()), verbosity=2)
        if self.iscolor():
            if colormap == 'gray':
                imshow(self.clone().grayscale().rgb()._array, figure=figure)
            else:
                imshow(self.clone().rgb()._array, figure=figure)
        else:
            imshow(self._array, colormap=colormap,
                   figure=figure, do_updateplot=flip)
        return self

    def imagesc(self):
        imshow(self.clone().mat2gray().rgb()._array)
        return self

    def mediatype(self):
        return 'image'

    def iscolor(self):
        return self.load().ndim == 3

    def filesize(self):
        assert self.hasfilename(), 'Invalid image filename'
        return os.path.getsize(self._filename)

    def width(self):
        return self.load().shape[1]

    def height(self):
        return self.load().shape[0]

    def shape(self):
        return self.tonumpy().shape

    def array(self, data=None):
        if data is None:
            return self._array
        elif isnumpy(data):
            self._array = np.copy(data)
            self._filename = None
            self._url = None
            return self
        else:
            raise ValueError('Invalid numpy array')

    def buffer(self, data=None):
        return self.array(data)

    def tonumpy(self):
        return self.array()

    def numpy(self):
        return self.tonumpy()

    def pil(self):
        return PIL.Image.fromarray(self.tonumpy())

    def filename(self, newfile=None):
        """Image Filename"""
        if newfile is None:
            return self._filename
        else:
            # set filename and return object
            self.flush()
            self._filename = newfile
            self._url = None
            return self

    def url(self, url=None, username=None, password=None, sha1=None, ignoreUrlErrors=None):
        """Image URL and URL download properties"""
        if url is None and username is None and password is None:
            return self._url
        if url is not None:
            # set filename and return object
            self.flush()
            self._filename = None
            self._url = url
        if username is not None:
            self._urluser = username  # basic authentication
        if password is not None:
            self._urlpassword = password  # basic authentication
        if sha1 is not None:
            self._urlsha1 = sha1  # file integrity
        if ignoreUrlErrors is not None:
            self._ignoreErrors = ignoreUrlErrors
        return self

    def uri(self):
        """Return the URI of the image object, either the URL or the filename, raise exception if neither defined"""
        if self._url is not None:
            return self._url
        elif self._filename is not None:
            return self._filename
        else:
            raise ValueError('No URI defined')

    def saveas(self, filename, writeas=None):
        if self.load().ndim == 3:
            imwrite(self.load(), filename, writeas=writeas)
        else:
            imwritegray(self.grayscale()._array, filename)
        self.flush()
        self._filename = filename
        return self

    def savetmp(self):
        return self.saveas(tempjpg())

    def savefig(self, filename=None):
        f = filename if filename is not None else tmpjpg()
        savefig(filename=f)
        return f

    def setattribute(self, key, value):
        if self.attributes is None:
            self.attributes = {key: value}
        else:
            self.attributes[key] = value
        return self

    def getattribute(self, key):
        if self.attributes is not None and key in list(self.attributes.keys()):
            return self.attributes[key]
        else:
            return None

    def hasattribute(self, key):
        return self.attributes is not None and key in self.attributes

    def hasurl(self):
        return self._url is not None and isurl(self._url)

    def hasfilename(self):
        return self._filename is not None and os.path.exists(self._filename)


    def stats(self):
        x = self._array.flatten()
        print('Channels: %d' % len(self._array))
        print('Shape: %s' % str(self._array.shape))
        print('min: %d' % np.min(x))
        print('max: %d' % np.max(x))
        print('mean: %d' % np.mean(x))
        print('std: %d' % np.std(x))

    # MODIFY IMAGE ---------------------------------------------------------
    def clone(self):
        """Create deep copy of image object"""
        im = copy.deepcopy(self)
        if self._array is not None:
            im._array = self._array.copy()
        return im

    def resize(self, cols=None, rows=None):
        """Resize the image buffer to (rows x cols)"""
        if cols is None or rows is None:
            if cols is None:
                scale = float(rows)/float(self.height())
            else:
                scale = float(cols)/float(self.width())
            #quietprint('[vipy.image][%s]: scale=%1.2f' %
            #           (self.__repr__(), scale), verbosity=2)

            # OpenCV decimation introduces artifacts using cubic
            # interp, INTER_AREA is recommended according to the
            # OpenCV docs
            #interp_method = cv2.INTER_AREA if scale < 1.0 else cv2.INTER_CUBIC            
            #self._array = cv2.resize(self.load(), dsize=(0, 0),
            #                       fx=scale, fy=scale,
            #                       interpolation=interp_method)
            self._array = imrescale(self.load(), scale)
            
        else:
            #quietprint('[vipy.image][%s]: resize=(%d,%d)' %
            #           (self.__repr__(), rows, cols), verbosity=2)
            try:
                #interp_method = cv2.INTER_AREA if (
                #    rows < self.height() or
                #    cols < self.width()) else cv2.INTER_CUBIC
                # fixed bug since opencv takes x and y not rows, cols
                #self._array = cv2.resize(self.load(), dsize=(cols, rows),
                #                       interpolation=interp_method)
                self._array = imresize(self.load(), rows, cols)

            except:
                print(self)  # DEBUGGING
                raise

        dtype = self._array.dtype
        if dtype == np.float32 or dtype == np.float64:
            np.clip(self._array, 0.0, 1.0, out=self._array)
        return self

    def rescale(self, scale=1):
        """Scale the image buffer by the given factor - NOT idemponent"""
        #quietprint('[vipy.image][%s]: scale=%1.2f to (%d,%d)' %
        #           (self.__repr__(), scale, scale*self.width(),
        #            scale*self.height()), verbosity=2)

        # OpenCV decimation introduces artifacts using cubic interp ,
        # INTER_AREA is recommended according to the OpenCV docs
        #interp_method = cv2.INTER_AREA if scale < 1.0 else cv2.INTER_CUBIC
        #self._array = cv2.resize(self.load(), dsize=(0, 0),
        #                       fx=scale, fy=scale, interpolation=interp_method)
        self._array = imrescale(self.load(), scale)

        dtype = self._array.dtype
        if dtype == np.float32 or dtype == np.float64:
            np.clip(self._array, 0.0, 1.0, out=self._array)
        return self

    def maxdim(self, dim):
        return self.rescale(float(dim) / float(np.maximum(self.height(), self.width())))

    def mindim(self, dim):
        return self.rescale(float(dim) / float(np.minimum(self.height(), self.width())))

    def pad(self, dx, dy, mode='edge'):
        """Pad image using np.pad mode"""
        self._array = np.pad(self.load(),
                           ((dx, dx), (dy, dy), (0, 0)) if
                           self.load().ndim == 3 else ((dx, dx), (dy, dy)),
                           mode=mode)
        return self

    def zeropad(self, dx, dy):
        """Pad image using np.pad constant"""
        if self.load().ndim == 3:
            pad_width = ((dx, dx), (dy, dy), (0, 0))
        else:
            pad_width = ((dx, dx), (dy, dy))
        self._array = np.pad(self.load(),
                           pad_width=pad_width,
                           mode='constant',
                           constant_values=0)        
        return self

    def meanpad(self, dx, dy):
        """Pad image using np.pad constant where constant is image mean"""
        #mu = self.mean()
        if self.load().ndim == 3:
            pad_size = ((dx, dx), (dy, dy), (0, 0))
            #constant_values = tuple([(x, y) for (x, y) in zip(mu, mu)])
        else:
            #constant_values = ((mu, mu), (mu, mu))
            pad_size = ((dx, dx), (dy, dy))
        self._array = np.pad(self.load(), pad_size, mode='mean')
        return self


    def minsquare(self):
        """Crop image of size (HxW) to (min(H,W), min(H,W))"""
        img = self.load()
        S = np.min(img.shape[0:2])
        self._array = self._array[0:S,0:S,:]
        return self

    def maxsquare(self):
        """Crop image of size (HxW) to (max(H,W), max(H,W)) with zeropadding"""
        img = self.load()
        S = np.max(img.shape)
        self._array = np.pad(self.load(), ((0,S-self.height()), (0,S-self.width()), (0,0)), mode='constant')
        self._array = self._array[0:S,0:S,:]
        return self

    def maxsquare_with_meanpad(self):
        """Crop image of size (HxW) to (max(H,W), max(H,W)) with zeropadding"""
        img = self.load()
        S = np.max(img.shape)
        self._array = np.pad(self.load(), ((0,S-self.height()), (0,S-self.width()), (0,0)), mode='mean')
        self._array = self._array[0:S,0:S,:]
        return self

    def crop(self, bbox=None, pad='mean'):
        """Crop the image buffer using the supplied bounding box - NOT
        idemponent."""
        if bbox is not None:
            if islist(bbox):
                bbox = BoundingBox(xmin=bbox[0], ymin=bbox[1],
                                   xmax=bbox[2], ymax=bbox[3])

            bbox = bbox.imclip(self.load())  # FIXME
            #quietprint('[vipy.image][%s]: cropping "%s"' %
            #           (self.__repr__(), str(bbox)), verbosity=2)
            bbox = bbox.imclip(self.load())
            # assumed numpy
            self._array = self.load()[int(bbox.ymin):int(bbox.ymax),
                                    int(bbox.xmin):int(bbox.xmax)]
        return self

    def fliplr(self):
        """Mirror the image buffer about the vertical axis - Not idemponent"""
        quietprint('[vipy.image][%s]: fliplr' %
                   (self.__repr__()), verbosity=2)
        self._array = np.fliplr(self.load())
        return self


    def normalize(self):
        """Convert image to float32 with [min,max] = [0,1]"""
        self._array = (self.load().astype(np.float32) - float(self.min())) / float(self.max()-self.min())
        return self

    def raw(self, normalized=True):
        """Load the image as a raw image buffer"""
        quietprint('[vipy.image][%s]: loading raw imagery data' %
                   (self.__repr__()), verbosity=2)
        self._array = self.load(asRaw=True)
        return self

    def grayscale(self):
        """Convert the image buffer to grayscale"""
        if self.load().ndim == 3:
            #quietprint('[vipy.image][%s]: converting to grayscale' %
            #           (self.__repr__()), verbosity=3)
            #self._array = cv2.cvtColor(self.load(), cv2.COLOR_BGR2GRAY)
            self._array = bgr2gray(self.load())
            self.setattribute('colorspace', 'gray')
        return self

    def greyscale(self):
        """Convert the image buffer to grayscale"""
        return self.grayscale()

    def rgb(self):
        """Convert the image buffer to RGB"""
        if self.load().ndim == 3:
            #quietprint('[vipy.image][%s]: converting bgr to rgb' %
            #           (self.__repr__()), verbosity=2)
            # opencv BGR to RGB
            #self._array = cv2.cvtColor(self.load(), cv2.COLOR_BGR2RGB)
            self._array = bgr2rgb(self.load())
        elif self.load().ndim == 2:
            #quietprint('[vipy.image][%s]: converting gray to rgb' %
            #           (self.__repr__()), verbosity=2)
            #self._array = cv2.cvtColor(self.load(), cv2.COLOR_GRAY2RGB)
            self._array = gray2rgb(self.load())
        self.setattribute('colorspace', 'rgb')
        return self

    def hsv(self):
        """Convert the image buffer to HSV color space"""
        if self.iscolor():
            #quietprint('[vipy.image][%s]: converting to hsv' %
            #           (self.__repr__()), verbosity=2)
            # opencv BGR (assumed) to HSV
            #self._array = cv2.cvtColor(self.load(), cv2.COLOR_BGR2HSV)
            self._array = bgr2hsv(self.load())
            self.setattribute('colorspace', 'hsv')
        else:
            #quietprint('[vipy.image][%s]: converting grayscale to hsv' %
            #           (self.__repr__()), verbosity=2)
            # grayscale -> RGB -> HSV (HACK)
            #self._array = cv2.cvtColor(self.rgb().load(), cv2.COLOR_RGB2HSV)
            self._array = gray2hsv(self.load())
        return self

    def bgr(self):
        """Convert the image buffer to BGR color format"""
        if self.load().ndim == 3:
            # Loaded by opencv, so assumed BGR order
            #self._array = self.load()
            #self._array = cv2.cvtColor(self.load(), cv2.COLOR_RGB2BGR)
            self._array = self.load()[:,:,::-1]
        elif self.load().ndim == 2:
            #quietprint('[vipy.image][%s]: converting gray to bgr' %
            #           (self.__repr__()), verbosity=2)
            #self._array = cv2.cvtColor(self.load(), cv2.COLOR_GRAY2BGR)
            self._array = gray2bgr(self.load())
        self.setattribute('colorspace', 'bgr')
        return self

    def float(self, scale=None):
        """Convert the image buffer to float32"""
        if self.load().dtype != np.float32:
            #quietprint('[vipy.image][%s]: converting to float32' %
            #           (self.__repr__()), verbosity=2)
            self._array = np.float32(self.load())
        if scale is not None:
            self._array = self._array * scale
        return self

    def uint8(self, scale=None):
        """Convert the image buffer to uint8"""
        if scale is not None:
            self._array = self.load() * scale
        if self.load().dtype != np.uint8:
            #quietprint('[vipy.image][%s]: converting to uint8' %
            #           (self.__repr__()), verbosity=2)
            self._array = np.uint8(self.load())
        return self

    def preprocess(self, scale=1.0/255.0):
        """Preprocess the image buffer by converting to grayscale, convert to
        float, then rescaling [0,255], [0,1] - Not idemponent."""
        quietprint('[vipy.image][%s]: preprocessing' %
                   (self.__repr__()), verbosity=2)
        self = self.grayscale().float()
        self._array = scale*self._array
        return self

    def min(self):
        return np.min(self.load().flatten())

    def max(self):
        return np.max(self.load().flatten())

    def mean(self):
        return np.mean(self.load(), axis=(0, 1)).flatten()

    def mat2gray(self, min=None, max=None):
        """Convert the image buffer so that [min,max] -> [0,1]"""
        quietprint('[vipy.image][%s]: contrast equalization' %
                   (self.__repr__()), verbosity=2)
        self._array = mat2gray(np.float32(self.load()), min, max)
        return self

    def transform2D(self, txy=(0, 0), r=0, s=1):
        """Transform the image buffer using a 2D similarity transform - Not
        idemponent."""
        quietprint('[vipy.image][%s]: transform2D' %
                   (self.__repr__()), verbosity=2)
        self.load()
        c = (self._array.shape[1] / 2, self._array.shape[0] / 2)
        M = similarity_imtransform2D(c=c, r=r, s=s)
        self._array = imtransform2D(self._array, M)
        A = similarity_imtransform(txy=txy)
        self._array = imtransform(self._array, A)
        return self

    def transform(self, A):
        """Transform the image buffer using the supplied affine transformation
        - Not idemponent."""
        quietprint('[vipy.image][%s]: transform' %
                   (self.__repr__()), verbosity=2)
        self._array = imtransform(self.load(), A)
        return self

    def gain(self, g):
        self._array = np.multiply(self.load(), g)
        return self

    def bias(self, b):
        self._array = self.load() + b
        return self

    def imrange(self):
        self._array = np.minimum(np.maximum(self.load(), 0), 255)
        return self

    def map(self, f):
        """Transform the image object using the supplied function - May not be
        idemponent."""
        return f(self)

    def drawbox(self, bbox, border=None, color=None, alpha=None, beta=None):
        self.load()
        dtype = self._array.dtype

        border = 2 if border is None else border
        alpha = 1.5 if alpha is None else alpha
        beta = 0.10 if dtype == np.float32 and beta is None else beta
        beta = 15 if beta is None else beta

        xmin = int(round(max(0, bbox.xmin-border)))
        ymin = int(round(max(0, bbox.ymin-border)))
        xmax = int(round(min(self.width(), bbox.xmax+border)))
        ymax = int(round(min(self.height(), bbox.ymax+border)))

        if dtype == np.float32:
            color = color if color is not None else 0.0
            clip = 1.0
        else:
            color = color if color is not None else (0, 196, 0)
            clip = 255

        data = self._array.astype(np.float32)
        data[ymin:ymax, xmin:xmax] *= alpha
        data[ymin:ymax, xmin:xmax] += beta

        data = np.clip(data, 0, clip)
        data = data.astype(dtype)

        self._array[ymin:ymin+border, xmin:xmax] = color
        self._array[ymax-border:ymax, xmin:xmax] = color
        self._array[ymin:ymax, xmin:xmin+border] = color
        self._array[ymin:ymax, xmax-border:xmax] = color

        return self

    def html(self, alt=None):
        im = self.clone().rgb()
        #ret, data = cv2.imencode('.png', im._array)
        #b = data.tobytes().encode('base64')

        buf = io.BytesIO()
        PIL.Image.frombuffer(im).save(buf, format='JPEG')
        b = buf.getvalue().encpde('base64')

        alt_text = alt if alt is not None else im.filename()
        return '<img src="data:image/png;base64,%s" alt="%s" />' % (b,
                                                                    alt_text)


class ImageCategory(Image):
    def __init__(self, filename=None, url=None, category=None,
                 attributes=None, array=None):
        # Image class inheritance
        super(ImageCategory, self).__init__(filename=filename,
                                            url=url,
                                            attributes=attributes,
                                            array=array)
        self._category = category

    def __repr__(self):
        if self.isloaded():
            str_size = ", height=%d, width=%d, color='%s'" % (
                self._array.shape[0], self._array.shape[1],
                str(self.getattribute('colorspace')))
        else:
            str_size = ""
        if self._url is None:
            if self._filename is None:
                str_file = ""
            else:
                str_file = "filename='%s'" % str(self._filename)
        else:
            str_file = "url='%s', filename='%s'" % (
                str(self._url), str(self._filename))

        str_category = ", category='%s'" % self._category
        return str('<vipy.imagecategory: %s%s%s>' % (str_file,
                                                      str_category, str_size))

    def __eq__(self, other):
        return self._category.lower() == other._category.lower()

    def __ne__(self, other):
        return self._category.lower() != other._category.lower()

    def is_(self, other):
        return self.__eq__(other)

    def is_not(self, other):
        return self.__ne__(other)

    def __hash__(self):
        return hash(self._category.lower())

    def iscategory(self, category):
        return (self._category.lower() == category.lower())

    def ascategory(self, newcategory):
        return self.category(newcategory)

    def category(self, newcategory=None):
        if newcategory is None:
            return self._category
        else:
            self._category = newcategory
            return self

    def score(self, newscore=None):
        """Real valued score for categorization, larger is better"""
        if newscore is None:
            return self.getattribute('score')
        else:
            self.setattribute('score', newscore)
            return self

    def probability(self, newprob=None):
        """Real valued probability for categorization, [0,1]"""
        if newprob is None:
            return self.getattribute('probability')
        else:
            self.setattribute('probability', newprob)
            self.setattribute('RawDetectionProbability', newprob)
            return self


class ImageDetection(ImageCategory):
    def __init__(self, filename=None, url=None, category=None,
                 attributes=None, xmin=float('nan'),
                 xmax=float('nan'), ymin=float('nan'), ymax=float('nan'), width=None, height=None,
                 bbox=None):
        # ImageCategory class inheritance
        super(ImageDetection, self).__init__(filename=filename,
                                             url=url,
                                             attributes=attributes,
                                             category=category)

        if bbox is not None:
            self.bbox = bbox
        elif width is None and height is None:
            self.bbox = BoundingBox(xmin=float(xmin), ymin=float(ymin),
                                    xmax=float(xmax), ymax=float(ymax))
        elif width is not None and height is not None:
            self.bbox = BoundingBox(xmin=float(xmin), ymin=float(ymin),
                                    width=float(width), height=float(height))
        else:
            raise ValueError('invalid parameterization')
            
    def __repr__(self):
        if self.isloaded():
            str_size = ", height=%d, width=%d, color='%s'" % (
                self._array.shape[0],
                self._array.shape[1],
                self.getattribute('colorspace'))
        else:
            str_size = ""
        str_category = "category='%s'" % self.category()
        str_detection = (", bbox=(xmin=%1.1f,ymin=%1.1f,"
                         "xmax=%1.1f,ymax=%1.1f)" % (
                             self.bbox.xmin, self.bbox.ymin,
                             self.bbox.xmax, self.bbox.ymax))
        str_url = "url='%s', " % str(self._url)
        if self._url is not None:
            str_file = ", url='%s', filename='%s'" % (str(self._url),
                                                      str(self._filename))
        else:
            if self.filename is None:
                str_file = ''
            else:
                str_file = ", filename='%s'" % (str(self._filename))

        return str('<vipy.imagedetection: %s%s%s%s>' % (
            str_category, str_detection, str_file, str_size))

    def __hash__(self):
        return hash(self.__repr__())

    def show(self, ignoreErrors=False, colormap=None, figure=None, flip=True):
        if self.load(ignoreErrors=ignoreErrors) is not None:
            if self.bbox.valid() and self.bbox.shape() != self._array.shape[0:2]:
                imbbox(self.clone().rgb()._array, self.bbox.xmin,
                       self.bbox.ymin, self.bbox.xmax, self.bbox.ymax,
                       bboxcaption=self.category(), colormap=colormap,
                       figure=figure, do_updateplot=flip)
            else:
                super(ImageDetection, self).show(figure=figure,
                                                 colormap=colormap, flip=flip)
        return self

    def boundingbox(self, xmin=None, xmax=None, ymin=None, ymax=None,
                    bbox=None, width=None, height=None, dilate=None,
                    xcentroid=None, ycentroid=None, dilate_topheight=None):
        if xmin is None and xmax is None and \
           ymin is None and ymax is None and \
           bbox is None and \
           width is None and height is None and \
           dilate is None and dilate_topheight is None:
            return self.bbox
        elif (xmin is not None and xmax is not None and
              ymin is not None and ymax is not None):
            self.bbox = BoundingBox(xmin=float(xmin), ymin=float(ymin),
                                    xmax=float(xmax), ymax=float(ymax))
        elif bbox is not None:
            self.bbox = bbox
        elif (xmin is not None and ymin is not None and
              width is not None and height is not None):
            try:
                self.bbox = BoundingBox(xmin=float(xmin), ymin=float(ymin),
                                        width=float(width),
                                        height=float(height))
            except ValueError as e:
                print(('{}: xmin={}, ymin={}, width={}, height={}'.format(
                    e, xmin, ymin, width, height)))
                raise e
        elif (xcentroid is not None and ycentroid is not None and
              width is not None and height is not None):
            self.bbox = BoundingBox(xcentroid=float(xcentroid),
                                    ycentroid=float(ycentroid),
                                    width=float(width),
                                    height=float(height))
        elif (dilate is None and dilate_topheight is None):
            raise ValueError('Invalid bounding box - Either rect coordinates '
                             'or bbox object must be provided')

        if dilate_topheight is not None:
            self.bbox.dilate_topheight(dilate_topheight)
        if dilate is not None:
            self.bbox.dilate(dilate)
        return self

    def drawbox(self, bbox=None, border=None, color=None, alpha=None,
                beta=None):
        bbox = bbox if bbox is not None else self.boundingbox()
        super(ImageDetection, self).drawbox(bbox=bbox, border=border,
                                            color=color,
                                            alpha=alpha, beta=beta)
        return self

    def invalid(self, flush=False):
        return self.bbox.invalid() or not super(ImageDetection, self).isvalid(
            flush=flush)

    def isinterior(self, W=None, H=None, flush=False):
        """Is the bounding box fully within the image rectangle?  Use provided
        image width and height (W,H) to avoid lots of reloads in some
        conditions"""
        (W, H) = (W if W is not None else self.width(),
                  H if H is not None else self.height())
        if flush:
            self.flush()
        return (self.bbox.xmin >= 0 and self.bbox.ymin >= 0 and
                self.bbox.xmax < W and self.bbox.ymax < H)

    def rescale(self, scale=1):
        """Rescale image buffer and bounding box - Not idemponent"""
        # image class inheritance
        self = super(ImageDetection, self).rescale(scale)
        self.bbox = self.bbox.rescale(scale)
        return self


    def resize(self, cols=None, rows=None):
        """Resize image buffer and bounding box"""
        # image class inheritance
        self = super(ImageDetection, self).resize(cols, rows)
        self.bbox = BoundingBox(xmin=0, ymin=0,
                                xmax=self.width(), ymax=self.height())
        return self

    def fliplr(self):
        """Mirror buffer and bounding box around vertical axis"""
        self = super(ImageDetection, self).fliplr()  # image class inheritance
        xmin = self.bbox.xmin
        xmax = self.bbox.xmax
        self.bbox.xmin = self.width() - xmax
        self.bbox.xmax = self.bbox.xmin + (xmax-xmin)
        return self

    def crop(self, bbox=None):
        """Crop image and update bounding box"""
        # image class inheritance
        self = super(ImageDetection, self).crop(
            bbox=self.bbox if bbox is None else bbox)
        if self.bbox is not None:
            self.bbox = BoundingBox(xmin=0, ymin=0,
                                    xmax=self.width(), ymax=self.height())
        return self

    def centercrop(self, bbwidth, bbheight):
        (W,H) = (self.width(), self.height())
        self.bbox = BoundingBox(xcentroid=W/2.0, ycentroid=H/2.0, width=bbwidth, height=bbheight)
        return self.crop()

    def clip(self, border=0):
        self.bbox = self.bbox.intersection(
            BoundingBox(
                xmin=border,
                ymin=border,
                xmax=self.width()-border,
                ymax=self.height()-border))
        return self

    def maxsquare(self):
        """Return a square bounding box centered at current centroid"""
        self.bbox = self.bbox.maxsquare()
        return self

    def dilate(self, s):
        """Dilate bounding box by scale factor"""
        self.bbox = self.bbox.dilate(s)
        return self

    def pad(self, dx, dy, mode='edge'):
        """Pad image using np.pad mode"""
        # image class inheritance
        self = super(ImageDetection, self).pad(dx, dy, mode)
        self.bbox = self.bbox.translate(dx, dy)
        return self

    def zeropad(self, dx, dy):
        """Pad image using np.pad constant"""
        # image class inheritance
        self = super(ImageDetection, self).zeropad(dx, dy)
        self.bbox = self.bbox.translate(dx, dy)
        return self

    def meanpad(self, dx, dy):
        """Pad image using np.pad constant where constant is the RGB mean per
        image"""
        # image class inheritance
        self = super(ImageDetection, self).meanpad(dx, dy)
        self.bbox = self.bbox.translate(dx, dy)
        return self

    def mask(self, W=None, H=None):
        """Return a binary array of the same size as the image (or using the
        provided image width and height (W,H) size to avoid an image load),
        with ones inside the bounding box"""
        if (W is None or H is None):
            (H, W) = (int(np.round(self.height())),
                      int(np.round(self.width())))
        immask = np.zeros((H, W)).astype(np.uint8)
        (ymin, ymax, xmin, xmax) = (int(np.round(self.bbox.ymin)),
                                    int(np.round(self.bbox.ymax)),
                                    int(np.round(self.bbox.xmin)),
                                    int(np.round(self.bbox.xmax)))
        immask[ymin:ymax, xmin:xmax] = 1
        return immask

    def setzero(self, bbox=None):
        """Set all image values within the bounding box to zero"""
        bbox = self.bbox if bbox is None else bbox
        self.load()[int(bbox.ymin):int(bbox.ymax),
                    int(bbox.xmin):int(bbox.xmax)] = 0
        return self



import os
import csv
from bobo.cache import Cache
from bobo.util import remkdir, isstring
from bobo.viset.stream import ImageCategoryStream
from nltk.corpus import wordnet

URL = 'http://image-net.org/imagenet_data/urls/imagenet_fall11_urls.tgz'
SHA1 = 'f5fd118232b871727fe333778be81df6c6fec372'
TXTFILE = 'fall11_urls.txt'
VISET = 'imagenet_fall2011'

cache = Cache()

def category(wnid):
    pos = wnid[0]
    synset = wordnet._synset_from_pos_and_offset(pos, int(str(wnid[1:]).lstrip('0')))  # assume noun
    return str(synset.lemmas[0].name).replace(" ","_")

def stream():
    csvfile = os.path.join(cache.root(), '%s.csv' % VISET)            
    if not os.path.isfile(csvfile):
        csvfile = export()        
    return ImageCategoryStream(csvfile, ignore=True)
         
def export():
    # Fetch textfile for construction
    txtfile = os.path.join(cache.root(), VISET, TXTFILE)
    if not os.path.isfile(txtfile):
        cache.unpack(cache.get(URL, key='imagenet_fall11_urls.tgz'), unpackto=os.path.join(cache.root(), VISET), sha1=SHA1, cleanup=True)
    outfile = os.path.join(cache.root(), '%s.csv' % VISET)
            
    # Write images and annotations
    impattern = os.path.join(VISET, '%08d.jpg')
    if not os.path.isfile(outfile):
        with open(outfile, 'wb') as csvfile:            
            f = csv.writer(csvfile, delimiter=' ', quotechar='|', quoting=csv.QUOTE_MINIMAL)        
            for (k, line) in enumerate(open(txtfile,'r')):
                try:
                    (name, url) = line.decode('utf-8').rstrip().split('\t')
                    (wnid, suffix) = name.decode('utf-8').rstrip().split('_')      
                    f.writerow([url.encode('utf-8'), impattern % k, wnid.encode('utf-8')])
                except KeyboardInterrupt:
                    raise
                except ValueError:
                    print '[viset.imagenet]: Warning: Ignoring malformed line "' + line[0:64] + ' ..."'                
                except:
                    raise

                if (k % 10000) == 0:
                    print '[bobo.viset.imagenet][%d/14200000]: exporting "%s"' % (k, url.encode('utf-8'))
    else:
        print '[bobo.viset.imagenet]: returning cached viset file "%s"' % (outfile)        

    # Cleanup
    if os.path.exists(txtfile):
        os.remove(txtfile)
        
    # Done!
    print outfile
    return outfile




import pygame
from PIL import Image
import multiprocessing
import sys
import os
import numpy as np
from bobo.util import bgr2rgb, numpy2iplimage
import cv2
import cv2.cv as cv
import signal
from time import sleep
import matplotlib.cm as cm 

raise ValueError('PyGame backend not supported')

# Window state
IMG = None
SCREEN = None
DRAWQUEUE = None
DRAWPROCESS = None

def _sigint_handler(signum, frame):
    sys.stdout.flush()
    pygame.quit()
    sys.exit()



# FIXME:can do one ctrl-c in ipython, but not two due to opencv and pygame reinitialization segfault
# this segfault makes pygame currently unusable as a general gui solution
signal.signal(signal.SIGINT, _sigint_handler)

def _drawqueue(funcname, argtuple):
    global DRAWQUEUE
    while DRAWQUEUE.empty() == False:
        sleep(0.01)  
    DRAWQUEUE.put((funcname, argtuple))
    
def imshow(im, title):
    figure()
    _drawqueue('_imshow', (im, title))

def rectangle(bbox, color='green', caption=None, filled=False, linewidth=1):
    figure()
    _drawqueue('_rectangle', (bbox, color, caption, filled, linewidth))

def ellipse(bbox, color='green', caption=None, filled=False, linewidth=1):
    figure()
    _drawqueue('_ellipse', (bbox, color, caption, filled, linewidth))
    
def circle(center, radius, color, caption, filled=False, linewidth=1):
    figure()
    _drawqueue('_circle', (center, radius, color, caption, filled, linewidth))


def frame(fr, im, color, caption):
    figure()
    _drawqueue('_frame', (fr, im, color, caption))

def tracks(im, bbox, bboxcolor, caption, captioncolor):
    figure()
    _drawqueue('_tracks', (im, bbox, bboxcolor, caption, captioncolor))
    
def scatter(fr, im, color):
    figure()
    _drawqueue('_scatter', (fr, im, color))

def figure(title=None):
    global DRAWPROCESS
    global DRAWQUEUE
    if (DRAWPROCESS is None) or (DRAWPROCESS.is_alive() == False):
        DRAWQUEUE = multiprocessing.Queue()
        p = multiprocessing.Process(target=_eventloop, args=(DRAWQUEUE,))
        #p.daemon = True
        p.start()        
        DRAWPROCESS = p
        
def close():
    global DRAWPROCESS
    if DRAWPROCESS is not None:
        pygame.display.quit()         
        DRAWPROCESS.terminate()
        DRAWPROCESS = None

def fullscreen():        
    figure()
    _drawqueue('_fullscreen', None)
    
    
def tracking(instream, framerate=None):
    # Initialize window
    pygame.init()
    (im, anno) = instream(async=True)[0]
    imsize = (im.get().shape[0], im.get().shape[1])
    screen = pygame.display.set_mode(imsize) 
    pygame.display.set_caption('tracking')
    imshown = im.url()
    img = pygame.image.load(im.url()) 

    # Initialize text
    font = pygame.font.SysFont(None, 12)

    # Update display
    for (im, anno) in instream(async=True):
        print anno        
        if anno['imurl'] != imshown:
            screen.blit(img, (0,0))
            if framerate is not None:
                pygame.time.wait(int(1000*(1.0/framerate)))
            pygame.display.flip() # update the display            
            img = pygame.image.load(im.url())             
            imshown = im.url()
        bbox = (anno['bbox_xmin'], anno['bbox_ymin'], anno['bbox_xmax']-anno['bbox_xmin'], anno['bbox_ymax']-anno['bbox_ymin'])
        pygame.draw.rect(img, (255,0,0), bbox, 1)

        text = font.render('%d' % anno['trackid'], 1, (0, 255, 0))
        textrect = text.get_rect()
        textrect.centerx = bbox[0] 
        textrect.centery = bbox[1]
        img.blit(text, textrect)
        
        #bbox(, imshape=(im.shape[0], im.shape[1]), bboxcaption=anno['trackid'])

def _pygame_to_cvimage(surface):
    """Convert a pygame surface into a cv image"""
    cv_image = cv.CreateImageHeader(surface.get_size(), cv.IPL_DEPTH_8U, 3)
    image_string = surface_to_string(surface)
    cv.SetData(cv_image, image_string)
    return cv_image
 
def _cvimage_to_pygame(image):
    """Convert cvimage into a pygame image"""
    image_rgb = cv.CreateMat(image.height, image.width, cv.CV_8UC3)
    cv.CvtColor(image, image_rgb, cv.CV_BGR2RGB)
    return pygame.image.frombuffer(image.tostring(), cv.GetSize(image_rgb), "RGB")
 
def _imshow(im, title=None, flip=True):
    global IMG
    global SCREEN
    
    if (type(im) is str) and os.path.isfile(im):
        imgfile = im
        im = Image.open(imgfile)  # do not load pixel buffer, just get size
        SCREEN = pygame.display.set_mode(im.size)         
        if title is not None:
            pygame.display.set_caption(title)
        IMG = pygame.image.load(imgfile) 
        SCREEN.blit(IMG, (0,0))
    else:
        if im.ndim == 2:
            im = cv2.cvtColor(im, cv2.COLOR_GRAY2RGB)            
            im = im.transpose(1,0,2)            
        else:
            im = im.transpose(1,0,2)
            im = im[:, :, ::-1]  # bgr -> rgb
        SCREEN = pygame.display.set_mode((im.shape[0], im.shape[1])) 
        pygame.surfarray.blit_array(SCREEN, np.uint8(im))
        if title is not None:
            pygame.display.set_caption(title)        
    if flip:
        pygame.display.flip() # update the display                        
    
def _circle(pos, radius, color='green', caption=None, filled=False, linewidth=1):    
    global IMG
    global SCREEN
    
    if filled:
        linewidth = 0
    pygame.draw.circle(SCREEN, pygame.Color(color), pos, radius, linewidth)
    if caption is not None:
        font = pygame.font.SysFont(None, 12)
        text = font.render('%s' % caption, 1, (0, 255, 0))
        textrect = text.get_rect()
        textrect.centerx = pos[0] 
        textrect.centery = pos[1]
        SCREEN.blit(text, textrect)        
    pygame.display.flip() # update the display                

    
def _rectangle(bbox, color='green', caption=None, captioncolor='red', filled=False, linewidth=1, flip=True):
    global IMG
    global SCREEN
    
    font = pygame.font.SysFont(None, 12)
    if filled:
        linewidth = 0
    pygame.draw.rect(SCREEN, pygame.Color(color), bbox, linewidth)
    #SCREEN.blit(IMG, (0,0))
    text = font.render('%s' % caption, 1, pygame.Color(captioncolor))
    textrect = text.get_rect()
    textrect.centerx = bbox[0] 
    textrect.centery = bbox[1]
    #IMG.blit(text, textrect)
    SCREEN.blit(text, textrect)    

    if flip:
        pygame.display.flip() # update the display                

def _ellipse(bbox, color='green', caption=None, filled=False, linewidth=1):
    global IMG
    global SCREEN
    
    font = pygame.font.SysFont(None, 12)
    if filled:
        linewidth = 0
    pygame.draw.ellipse(SCREEN, pygame.Color(color), bbox, linewidth)
    #SCREEN.blit(IMG, (0,0))
    text = font.render('%s' % caption, 1, (0, 255, 0))
    textrect = text.get_rect()
    textrect.centerx = bbox[0] 
    textrect.centery = bbox[1]
    #IMG.blit(text, textrect)
    SCREEN.blit(text, textrect)    
    
    pygame.display.flip() # update the display                

def _frame(fr, im=None, color='green', linewidth=1):
    global SCREEN

    _imshow(im, flip=False)
    for xysr in fr.transpose():
        bbox = (xysr[0], xysr[1], 10, 10)
        s = xysr[2]
        th = xysr[3]
        R = np.mat([[np.cos(th), -np.sin(th)], [np.sin(th), np.cos(th)]])
        x = np.mat([ [-10, +10, +10, -10], [-10, -10, +10, +10] ])
        y = R*((s/10)*x)
        #pygame.draw.rect(SCREEN, pygame.Color(color), bbox, 1)
        pygame.draw.aalines(SCREEN, pygame.Color(color), True, [(xysr[0]+y[0,0],xysr[1]+y[1,0]), (xysr[0]+y[0,1],xysr[1]+y[1,1]), (xysr[0]+y[0,2],xysr[1]+y[1,2]), (xysr[0]+y[0,3],xysr[1]+y[1,3])])  # for rotated square
    pygame.display.flip() # update the display                

    
def _tracks(im, bbox, bboxcolor, caption, captioncolor):
    global SCREEN

    _imshow(im, flip=False)
    for (bb,cap) in zip(bbox, caption):
        _rectangle(bb, bboxcolor, cap, captioncolor, flip=False)
    pygame.display.flip() # update the display                

    
def _flow(fr, uv_flow):
    pass
    
def _scatter(fr, im=None, color='green', linewidth=1):
    global SCREEN

    n_frame = fr.shape[1]
    cmap = cm.get_cmap('jet', n_frame) 
    rgb = np.uint8(255*cmap(np.arange(n_frame)))
    _imshow(im, flip=False)
    for (i,xysr) in enumerate(fr.transpose()):
        pygame.draw.circle(SCREEN, pygame.Color(int(rgb[i,0]), int(rgb[i,1]), int(rgb[i,2]), int(rgb[i,3])), (int(xysr[0]), int(xysr[1])), 1, 1)        
    pygame.display.flip() # update the display                

    
def _fullscreen():
    pygame.display.toggle_fullscreen()  # doesn't work

        
def _eventloop(drawqueue):
    pygame.init()
    imfile = os.path.join(os.path.dirname(__file__),'..','..','data','visym_owl.png')
    pygame.display.set_icon(pygame.image.load(imfile))

    # Initialize display
    global SCREEN
    SCREEN = pygame.display.set_mode((320, 240))         
    pygame.display.set_caption('Figure')
    pygame.display.flip() # update the display                
    
    # Event loop
    Clock = pygame.time.Clock()
    done = False
    while not done:
        # GUI events
        Clock.tick(60)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                done = True 
            elif event.type == pygame.KEYDOWN:
                if event.key == pygame.K_ESCAPE:
                    #done = True 
                    done = False

        # Drawing events
        if not drawqueue.empty():
            (funcname, args) = drawqueue.get()
            func = globals()[funcname]
            if args is not None:
                func(*args)
            else:
                func()
            
    sys.stdout.flush()
    pygame.quit()
    sys.exit()
    





from bobo.cache import Cache
from bobo.viset import ethzshapes

ethzshapes.URL = 'http://www.vision.ee.ethz.ch/datasets_extra/extended_ethz_shapes.tgz'
ethzshapes.SHA1 = None
ethzshapes.SUBDIR = 'extended_ethz_shapes'
ethzshapes.LABELS = ['apple','bottle','giraffe','hat','mug','starfish','swan']
ethzshapes.VISET = 'ethzshapes_extended'

cache = Cache()

def stream():
    return ethzshapes.stream()

def export():
    return ethzshapes.export()




"""Annotation tools for downloading videos from youtube"""

from vipy.annotation import common_user_agents
import os
import ssl
import urllib
import json
import re
import random
from vipy.util import tofilename, remkdir, filepath, filebase, isurl

def dailymotion(tag):
    pass
    
def vine(user_id, outdir=None, video_limit=None):
    # url = 'https://vine.co/u/%s/' % user_id
    url = 'https://api.vineapp.com/timelines/users/%s' % user_id

    user_agent = random.choice(common_user_agents)
    headers = {'User-Agent':user_agent}
    search_request = urllib.request.Request(url, None, headers)
    try:
        search_results = urllib.request.urlopen(search_request)
    except:
        print('[bobo.videosearch.vine]: Error on URL: %s' % (url))
        return ([None], [None])
    search_data = search_results.read()
    data = json.loads(search_data)
    try:
        # https://github.com/starlock/vino/wiki/API-Reference#response-6
        vidlist = [rec['videoUrl'] for rec in data['data']['records']]
    except (KeyError, what):
        print('[bobo.videosearch.vine]: Key error on parse: %s' % (what))
        return ([None], [None])

    vidlist = list(set(vidlist))  # unique
    if outdir is not None:
        remkdir(outdir)
        filelist = download(vidlist, os.path.join(outdir, 'vine_%s_%%04d.mp4' % user_id), video_limit=video_limit)
        return vidlist, filelist
    return(vidlist)

def isactiveyoutuber(username):
    """Does the youtube user have any uploaded videos?"""
    url = 'http://www.youtube.com/user/%s/videos' % username
    user_agent = random.choice(common_user_agents)
    headers = {'User-Agent':user_agent}
    search_request = urllib.request.Request(url, None, headers)
    try:
        gcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1) # to avoid [SSL: CERTIFICATE_VERIFY_FAILED] exception
        search_results = urllib.request.urlopen(search_request, context=gcontext)
    except AttributeError:
        try:
            search_results = urllib.request.urlopen(search_request)
        except:
            return False
    except:
        return False
    s = search_results.read()
    return 'Uploads' in str(s)
        
    
def youtubeuser(tag, n_pages=1):
    """return all unique /user/* urls returned for a search for a given query tag"""
    from BeautifulSoup import BeautifulSoup # venv
    url = 'http://www.youtube.com/results?search_query=%s&page=%d'
    userlist = []
    for k in range(0, n_pages):
        user_agent = random.choice(common_user_agents)
        headers = {'User-Agent':user_agent}
        search_request = urllib.request.Request(url % (tag.replace(' ','+'), k+1), None, headers)
        try:
            gcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1) # to avoid [SSL: CERTIFICATE_VERIFY_FAILED] exception
            search_results = urllib.request.urlopen(search_request, context=gcontext)
        except AttributeError:
            search_results = urllib.request.urlopen(search_request)
        links = BeautifulSoup(search_results.read()).findAll("a")
        for link in links:
            if len(link['href']) > 6 and '/user/' == link['href'][0:6]:
                userlist.append(str('http://www.youtube.com%s' % link['href']))
    return list(set(userlist))
    
        
def youtube(tag, n_pages=1, outdir=None, channel=False, video_limit=None, expected_vid_list=None):
    if channel:
        url = 'http://www.youtube.com/user/%s/videos'
        #url = 'http://www.youtube.com/user/%s/videos?sort=dd&view=0&flow=list&live_view=500'
    else:
        url = 'http://www.youtube.com/results?search_query=%s&page=%d'
    vidlist = []
    for k in range(0, n_pages):
        user_agent = random.choice(common_user_agents)
        headers = {'User-Agent':user_agent}
        if channel:
            search_request = urllib.request.Request(url % (tag.replace(' ','+')), None, headers)
        else:
            search_request = urllib.request.Request(url % (tag.replace(' ','+'), k+1), None, headers)
        try:
            gcontext = ssl.SSLContext(ssl.PROTOCOL_TLSv1) # to avoid [SSL: CERTIFICATE_VERIFY_FAILED] exception
            search_results = urllib.request.urlopen(search_request, context=gcontext)
        except AttributeError:
            search_results = urllib.request.urlopen(search_request)
        except:
            print('[bobo.videosearch.youtube]: URL 404: %s' % (url % (tag.replace(' ','+'))))
            url = 'http://www.youtube.com/channel/%s/videos'
            search_request = urllib.request.Request(url % (tag.replace(' ','+')), None, headers)
            try:
                search_results = urllib.request.urlopen(search_request, context=gcontext)
            except AttributeError:
                search_results = urllib.request.urlopen(search_request)
            except:
                print('[bobo.videosearch.youtube]: URL 404: %s' % (url % (tag.replace(' ','+'))))
                return ([None], [None])
        search_data = search_results.read()
        
        datalist = search_data.split('href="/watch?')

        vidlist.extend(['https://www.youtube.com/watch?%s' % vid.split('"')[0] for vid in datalist if 'DOCTYPE' not in vid.split('"')[0]])
        if expected_vid_list is not None:
            new_list = [v for v in expected_vid_list if v in vidlist]
            if len(new_list) == len(expected_vid_list):
                print('YOUTUBE: PREVIOUS USER CHECK WAS GOOD: %s' % tag)
                return new_list
            else:
                return []
    vidlist = list(set(vidlist))  # unique
    if outdir is not None:
        filelist = download(vidlist, os.path.join(remkdir(outdir), 'youtube_'+tofilename(tag, hyphen=False)+'_%04d.mp4'), video_limit=video_limit)
        #FIXME: Add this to others
        return vidlist, filelist
    return(vidlist)

def liveleak(tag, n_pages=1, outdir=None):
    url = 'http://www.liveleak.com/browse?q=%s&page=%d'
    vidlist = []
    for k in range(0, n_pages):
        user_agent = random.choice(common_user_agents)
        headers = {'User-Agent':user_agent}

        search_request = urllib.request.Request(url % (tag.replace(' ','+'), k+1), None, headers)
        search_results = urllib.request.urlopen(search_request)
        search_data = search_results.read()

        datalist = search_data.split('href="http://www.liveleak.com/view?')
        vidlist.extend(['http://www.liveleak.com/view?%s' % vid.split('"')[0].split('&')[0] for vid in datalist if 'DOCTYPE' not in vid.split('"')[0]])
    vidlist = list(set(vidlist))  # unique

    if outdir is not None:
        download(vidlist, os.path.join(remkdir(outdir), 'liveleak_'+tofilename(tag)+'_%04d.mp4'))
    return(vidlist)


def ustream(tag, n_pages=1, outdir=None):
    url = 'http://www.ustream.tv/search?q=%s'
    vidlist = []
    for k in range(0, 1):
        user_agent = random.choice(common_user_agents)
        headers = {'User-Agent':user_agent}

        search_request = urllib.request.Request(url % (tag.replace(' ','+')), None, headers)
        search_results = urllib.request.urlopen(search_request)
        search_data = search_results.read()

        datalist = search_data.split('href="/recorded/')
        vidlist.extend(['http://www.ustream.tv/recorded/%s' % vid.split('"')[0] for vid in datalist if 'DOCTYPE' not in vid.split('"')[0]])
    vidlist = list(set(vidlist))  # unique

    print(vidlist)
    if outdir is not None:
        download(vidlist, os.path.join(remkdir(outdir), 'ustream_'+tofilename(tag)+'_%04d.mp4'))
    return(vidlist)


def download(vidlist, outfile, skip=False, writeurlfiles=True, max_filesize='350m', video_limit=None, remove_parts=True):
    """Use youtube-dl to download videos from url"""
    vidfiles = []
    successes = 0
    for (k,v) in enumerate(vidlist):
        try:
            try:
                vidfile = outfile % k
            except TypeError:
                # Allow caller to explicitly name file
                vidfile = outfile
            print('[bobo.youtube.download]: exporting "%s" to "%s"' % (v, vidfile))
            erno = os.system('youtube-dl "%s" -o %s --max-filesize %s --no-check-certificate' % (v, vidfile, max_filesize))  # must be on path
            print('youtube-dl returned %d' % erno)
            if os.path.isfile(vidfile):
                vidfiles.append(vidfile)
                successes += 1
                if writeurlfiles:
                    urlfile = os.path.join(filepath(vidfile), '%s.url' % filebase(vidfile))
                    with open(urlfile, 'w') as f:
                        f.write(v + '\n')
            elif remove_parts and os.path.isfile(vidfile + '.part'):
                partfile = vidfile + '.part'
                print('[bobo.youtube.download]: removing partial file: %s' % partfile)
                os.remove(partfile)
            if video_limit is not None and successes == video_limit:
                break
        except Exception as exception:
            print(exception)
            # http://rg3.github.io/youtube-dl/supportedsites.html
            print('[bobo.youtube.download]: download failed - skipping')
        else:
            if skip and erno: #use this if every video on a page fails
                break
    return vidfiles



import os
from vipy.util import isnumpy, quietprint, isstring, isvideo, tempcsv, imlist, remkdir, filepath, filebase
from vipy.image import Image, ImageCategory, ImageDetection
import copy
import numpy as np

class Video(object):
    """Requires FFMPEG on the path for video files"""
    def __init__(self, frames=None, filenames=None, frameDirectory=None, attributes=None, startframe=0, videofile=None, rot90clockwise=False, rot270clockwise=False, framerate=2):
        if frames is not None:
            self._framelist = frames
        elif videofile is not None:
            imdir = remkdir(os.path.join(filepath(videofile), '.'+filebase(videofile)), flush=True)
            impattern = os.path.join(imdir, '%08d.png')
            if rot90clockwise:
                cmd = 'ffmpeg -i "%s" -vf "transpose=1, fps=%d" -f image2 "%s"' % (videofile, framerate, impattern)  # 90 degrees clockwise
            elif rot270clockwise:
                cmd = 'ffmpeg -i "%s" -vf "transpose=2, fps=%d" -f image2 "%s"' % (videofile, framerate, impattern)  # 270 degrees clockwise
            else:
                cmd = 'ffmpeg -i "%s" -vf fps=%d -f image2 "%s"' % (videofile, framerate, impattern)
            print('[VideoCategory]: Exporting frames using "%s" ' % cmd)
            os.system(cmd)  # HACK
            self._framelist = [Image(filename=imfile) for imfile in imlist(imdir)]
            self._videofile = videofile            
        elif filenames is not None:
            self._framelist = [Image(filename=imfile) for imfile in filenames]
        elif frameDirectory is not None:
            if not os.path.isdir(frameDirectory):
                raise ValueError('Invalid frames directory "%s"' % frameDirectory)
            self._framelist = [Image(filename=imfile) for imfile in imlist(frameDirectory)]
        else:
            self._framelist = []
            
        self._startframe = startframe

        # Public
        self.attributes = attributes
        
    def __repr__(self):
        return str('<strpy.video: frames=%d>' % (len(self._framelist)))
    
    def __iter__(self):
        for im in self._framelist[self._startframe:]:
            yield im
    
    def __len__(self):
        return len(self._framelist)
            
    def __getitem__(self, k):
        if k >= 0 and k < len(self._framelist) and len(self._framelist) > 0:
            return self._framelist[k]
        else:
            raise ValueError('Invalid frame index %d ' % k)

    def mediatype(self):
        return 'video'

    def append(self, im):
        self._framelist.append(im)
        return self
    
    def frames(self, newframes=None):
        if newframes is None:
            return self._framelist
        else:
            self._framelist = newframes
            return self

    def setattribute(self, key, value):
        if self.attributes is None:
            self.attributes = {key:value}
        else:
            self.attributes[key] = value
        return self

    def show(self, figure=1, colormap=None, do_flush=True):
        for im in self:  # use iterator
            im.show(figure=figure, colormap=colormap)
            #if do_flush:
            #    im.flush()  # flush after showing to avoid running out of memory
        return self
    
    def play(self, figure=1, colormap=None):
        return self.show(figure, colormap)
            
    def map(self, f):
        self._framelist = [f(im) for im in self._framelist]
        return self

    def filter(self, f):
        self._framelist = [im for im in self._framelist if f(im)]
        return self
    
    def clone(self):
        return copy.deepcopy(self)

    def flush(self):
        self._framelist = [im.flush() for im in self._framelist]
        return self

    def isvalid(self):
        return np.all([im.isvalid() for im in self._framelist])

    
class VideoCategory(Video):    
    def __init__(self, frames=None, filenames=None, frameDirectory=None, attributes=None, startframe=0, category=None, videofile=None, rot90clockwise=False, rot270clockwise=False, framerate=2):
        if frames is not None and len(frames) > 0:
            self._framelist = frames
            self._category = frames[0].category()
        elif videofile is not None:
            imdir = remkdir(os.path.join(filepath(videofile), '.'+filebase(videofile)), flush=True)
            impattern = os.path.join(imdir, '%08d.png')
            if rot90clockwise:
                cmd = 'ffmpeg -i "%s" -vf "transpose=1, fps=%d" -f image2 "%s"' % (videofile, framerate, impattern)  # 90 degrees clockwise
            elif rot270clockwise:
                cmd = 'ffmpeg -i "%s" -vf "transpose=2, fps=%d" -f image2 "%s"' % (videofile, framerate, impattern)  # 270 degrees clockwise
            else:
                cmd = 'ffmpeg -i "%s" -vf fps=%d -f image2 "%s"' % (videofile, framerate, impattern)
            print('[VideoCategory]: Exporting frames using "%s" ' % cmd)
            os.system(cmd)  # HACK
            self._framelist = [ImageCategory(filename=imfile, category=category) for imfile in imlist(imdir)]
            self._videofile = videofile
        elif filenames is not None and len(filenames) > 0:
            self._framelist = [ImageCategory(filename=imfile, category=category) for imfile in filenames]
        elif frameDirectory is not None:
            if not os.path.isdir(frameDirectory):
                raise ValueError('Invalid frames directory "%s"' % frameDirectory)
            self._framelist = [ImageCategory(filename=imfile, category=category) for imfile in imlist(frameDirectory)]
        else:
            self._framelist = []
            
        self._startframe = startframe

        if category is not None:
            self.category(category)
        
        # Public
        self.attributes = attributes
        
    def __repr__(self):
        return str('<strpy.videocategory: frames=%d, category="%s">' % (len(self), self._category))
                    
    def __eq__(self, other):
        return self._category.lower() == other._category.lower()

    def __ne__(self, other):
        return self._category.lower() != other._category.lower()

    def is_(self, other):
        return self.__eq__(other)

    def is_not(self, other):
        return self.__ne__(other)
            
    def __hash__(self):
        return hash(self._category.lower())                

    def iscategory(self, newcategory):
        return (self._category.lower() == newcategory.lower())

    def ascategory(self, newcategory):
        return self.category(newcategory)
    
    def category(self, newcategory=None):
        if newcategory is None:
            return self._category
        else:
            self._category = newcategory
            self.map(lambda im: im.category(newcategory))
            return self

    
class VideoDetection(VideoCategory):
    def __init__(self, frames=None, filenames=None, frameDirectory=None, attributes=None, startframe=0, category=None, boundingboxes=None, rectlist=None, videofile=None, rot90clockwise=False, rot270clockwise=False, framerate=2):
        # Public
        self.attributes = attributes
        
        if frames is not None and len(frames) > 0:
            self._framelist = frames
            self._category = frames[0].category()
            self.attributes = frames[0].attributes
        elif videofile is not None:
            imdir = remkdir(os.path.join(filepath(videofile), '.'+filebase(videofile)), flush=True)
            impattern = os.path.join(imdir, '%08d.png')
            if rot90clockwise:
                cmd = 'ffmpeg -i "%s" -vf "transpose=1, fps=%d" -f image2 "%s"' % (videofile, framerate, impattern)  # 90 degrees clockwise
            elif rot270clockwise:
                cmd = 'ffmpeg -i "%s" -vf "transpose=2, fps=%d" -f image2 "%s"' % (videofile, framerate, impattern)  # 270 degrees clockwise
            else:
                cmd = 'ffmpeg -i "%s" -vf fps=%d -f image2 "%s"' % (videofile, framerate, impattern)
            print('[VideoCategory]: Exporting frames using "%s" ' % cmd)
            os.system(cmd)  # HACK
            self._framelist = [ImageDetection(filename=imfile, category=category) for imfile in imlist(imdir)]
            self._videofile = videofile
            
        elif filenames is not None:
            self._framelist = [ImageDetection(filename=imfile, category=category) for imfile in filenames]
        elif frameDirectory is not None:
            if not os.path.isdir(frameDirectory):
                raise ValueError('Invalid frames directory "%s"' % frameDirectory)
            self._framelist = [ImageDetection(filename=imfile, category=category) for imfile in imlist(frameDirectory)]
            
        if boundingboxes is not None:
            self._framelist = [im.boundingbox(bb) for (im,bb) in zip(self._framelist, boundingboxes)]
        elif rectlist is not None:
            self._framelist = [im.boundingbox(xmin=r[0], ymin=r[1], xmax=r[2], ymax=r[3]) for (im,r) in zip(self._framelist, rectlist)]
            
        self._startframe = startframe

        if category is not None:
            self.category(category)
        
                                
    def __repr__(self):
        return str('<strpy.videodetection: frames=%d, category="%s">' % (len(self), str(self.category())))
    


class VideoCapture(object):
    """ Wraps OpenCV VideoCapture in a generator"""
    import cv2 
    def __init__(self, url, do_msec=False):
        """Takes a url, filename, or device id that produces video"""
        self.url = url
        self.do_msec = do_msec
        # Grab video metadata
        self.cap = cv2.VideoCapture(self.url)
        self.num_frames = self.cap.get(cv.CV_CAP_PROP_FRAME_COUNT)
        self.fps = self.cap.get(cv.CV_CAP_PROP_FPS)
        # Release file until we start iteration
        self.cap = self.cap.release()

    def __iter__(self):
        """Yields frames from the video source"""
        if self.cap is None:
            self.cap = cv2.VideoCapture(self.url)
            self.num_frames = self.cap.get(cv.CV_CAP_PROP_FRAME_COUNT)
            self.fps = self.cap.get(cv.CV_CAP_PROP_FPS)
        ret = True
        try:
            while ret:
                ret, frame = self.cap.read()
                if ret:
                    if self.do_msec:
                        msec = self.cap.get(cv.CV_CAP_PROP_POS_MSEC)
                        yield msec, frame
                    else:
                        yield frame
        finally:
            self.cap = self.cap.release()



import numpy as np
from vipy.util import load, save, csvlist, dirlist, imlist, readcsv, filebase, imwrite, remkdir, temppng, writecsv, filepath, fileext, filetail, islist
from vipy.image import ImageDetection, ImageCategory
import os
import copy


class ImageNet(object):
    def __init__(self, datadir='/proj/janus3/ILSVRC2015'):
        self.datadir = datadir 
        
    def __repr__(self):
        return str('<viset.imagenet: %s>' % self.datadir)

    def _parse_loc(self, imageset='train'):
        """ImageNet localization, imageset = {train, val}"""
        import xmltodict
        if imageset == 'train':
            imagesetfile = 'train_loc.txt'
        elif imageset == 'val':
            imagesetfile = 'val.txt'
        else:
            raise ValueError('unsupported imageset')

        csv = readcsv(os.path.join(self.datadir, 'ImageSets', 'CLS-LOC', imagesetfile), separator=' ')
        for (filepath, k) in csv:
            xmlfile = '%s.xml' % os.path.join(self.datadir, 'Annotations', 'CLS-LOC', imageset, filepath)
            d = xmltodict.parse(open(xmlfile, 'r').read())
            imfile = '%s.JPEG' % os.path.join(self.datadir, 'Data', 'CLS-LOC', imageset, filepath)
            objlist = d['annotation']['object'] if islist(d['annotation']['object']) else [d['annotation']['object']]
            for obj in objlist:
                yield ImageDetection(filename=imfile, category=obj['name'],
                                        xmin=int(obj['bndbox']['xmin']), ymin=int(obj['bndbox']['ymin']),
                                        xmax=int(obj['bndbox']['xmax']), ymax=int(obj['bndbox']['ymax']))

    def classes(self):
        return list(set([im.category() for im in self._parse_cls('val')]))

    def _parse_cls(self, imageset='train'):
        """ImageNet Classification, imageset = {train, val}"""
        import xmltodict            
        if imageset == 'train':
            imagesetfile = 'train_cls.txt'
        elif imageset == 'val':
            imagesetfile = 'val.txt'
        else:
            raise ValueError('unsupported imageset')
        csv = readcsv(os.path.join(self.datadir, 'ImageSets', 'CLS-LOC', imagesetfile), separator=' ')
        for (subpath, k) in csv:
            xmlfile = '%s.xml' % os.path.join(self.datadir, 'Annotations', 'CLS-LOC', imageset, subpath)
            imfile = '%s.JPEG' % os.path.join(self.datadir, 'Data', 'CLS-LOC', imageset, subpath)            
            if os.path.exists(xmlfile):
                d = xmltodict.parse(open(xmlfile, 'r').read())
                objlist = d['annotation']['object'] if islist(d['annotation']['object']) else [d['annotation']['object']]                
                yield ImageCategory(filename=imfile, category=objlist[0]['name'])
            else:
                yield ImageCategory(filename=imfile, category=filepath(subpath))
                

    



import os
import csv
from bobo.cache import Cache
from bobo.util import remkdir, isstring
from bobo.image import ImageDetectionStream

URL = 'http://tamaraberg.com/faceDataset/originalPics.tar.gz'
URL_ANNO = 'http://vis-www.cs.umass.edu/fddb/FDDB-folds.tgz'
VISET = 'fddb'

cache = Cache()




import numpy as np
from vipy.util import load, save, csvlist, txtlist, dirlist, imlist, readcsv, filebase, imwrite, remkdir, temppng, writecsv, filepath, loadmat
from vipy.image import ImageDetection, Image
import os
import copy


class VGGFaceURL(object):
    def __init__(self, datadir='/proj/janus3/vgg-face/curated/vgg_face_dataset'):
        self.datadir = datadir 
        self._subjects = None
        
    def __repr__(self):
        return str('<viset.vggfaceurl: %s>' % self.datadir)

    def subjects(self):
        if self._subjects is None:
            self._subjects = [filebase(f) for f in txtlist(os.path.join(self.datadir, 'data'))]
        return self._subjects # cached
    
    def dataset(self):
        """Return a generator to iterate over dataset"""  
        SCHEMA = ['id', 'url', 'left', 'top', 'right', 'bottom', 'pose', 'detection_score', 'curation']
        for f in txtlist(os.path.join(self.datadir, 'files')):
            for r in readcsv(f, separator=' '):
                im = ImageDetection(url=r[2], category=filebase(f), xmin=float(r[3]), ymin=float(r[4]), xmax=float(r[5]), ymax=float(r[6]),attributes=dict(zip(SCHEMA,r)))
                yield im

    def take(self, n):
        """Randomly select n frames from dataset"""
        takelist = []
        for csvfile in np.random.choice(txtlist(os.path.join(self.datadir, 'data')), n):
            csv = readcsv(csvfile, separator=' ')
            r = csv[np.random.randint(1,len(csv))] # not including header
            im = ImageDetection(url=r[2], category=filebase(csvfile), xmin=float(r[3]), ymin=float(r[4]), xmax=float(r[5]), ymax=float(r[6]), attributes=dict(zip(SCHEMA,r)))
            takelist.append(im)
        return takelist
                    
                                

    
class VGGFace(object):
    def __init__(self, datadir='/proj/janus3/vgg-face/vgg-face-janus'):
        self.datadir = datadir 
        self._subjects = None
        
    def __repr__(self):
        return str('<viset.vggface: %s>' % self.datadir)

    def subjects(self):
        if self._subjects is None:
            self._subjects = [filebase(d) for d in dirlist(os.path.join(self.datadir, 'images'))]
        return self._subjects # cached

    def wordnetid_to_name(self):
        wnid = [str(x[0]) for x in loadmat('/proj/janus3/vgg-face/vgg-face-janus/vgg_face_wordnetid.mat')['name'][0]]
        name = [str(x[0][0]) for x in loadmat('/proj/janus3/vgg-face/vgg-face-janus/vgg_face_subjects.mat')['subjects']]
        return {k:v for (k,v) in zip(wnid, name)}
        
    def dataset(self):
        """Return a generator to iterate over dataset"""
        for d in dirlist(os.path.join(self.datadir, 'images')):
            for f in imlist(d):
                im = ImageDetection(filename=f, category=filebase(d))
                im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                im = im.boundingbox(dilate=0.875)  # central 224x224
                yield im

    def fastset(self):
        """Return a generator to iterate over dataset"""
        for d in dirlist(os.path.join(self.datadir, 'images')):
            for f in imlist(d):
                im = ImageDetection(filename=f, category=filebase(d))
                #im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                #im = im.boundingbox(dilate=0.875)  # central 224x224
                yield im

    def take(self, n):
        S = np.random.choice(self.subjects(), n)
        takelist = []
        for d in dirlist(os.path.join(self.datadir, 'images')):
            if filebase(d) in S:
                f = np.random.choice(imlist(d),1)[0]
                im = ImageDetection(filename=f, category=filebase(d))                
                im = im.boundingbox(xmin=float(im.width()-256)/2.0, ymin=float(im.height()-256.0)/2.0, xmax=256.0+((im.width()-256.0)/2.0),ymax=256.0+((im.height()-256.0)/2.0))
                im = im.boundingbox(dilate=0.875)  # central 224x224
                takelist.append(im)
        return takelist

    def by_subject(self, wordnetid):
        d = os.path.join(self.datadir, 'images', wordnetid)
        for f in imlist(d):
            yield ImageDetection(filename=f, category=filebase(d))



import sys
sys.path.append('..')  # FIXME: relative import
from vipy.image import ImageDetection, Image, ImageCategory

def run():
    # Empty constructor should not raise exception
    im = Image()
    print('Empty Constructor: PASSED')

    # Non-existant filename should not raise exception during constructor (only during load)
    im = Image(filename='myfile')
    print('Filename Constructor: PASSED')

    # Malformed URL should raise exception
    try:
        im = Image(url='myurl')
        raise
    except:
        print('Malformed URL constructor: PASSED')

    # Valid URL should not raise exception (even if it is not an image extension)
    im = Image(url='http://visym.com')
    print('Image URL constructor: PASSED')

    # Valid URL and filename to save it
    im = Image(url='http://visym.com/myfile.jpg', filename='/my/file/path')
    print('URL and filename constructor: PASSED')

    # URL object
    im = Image(url='https://upload.wikimedia.org/wikipedia/commons/thumb/2/23/Bubo_virginianus_06.jpg/1920px-Bubo_virginianus_06.jpg')
    print('  Image __desc__: %s' % im)
    print('  Image length: %d' %  len(im))
    im.download()
    print('  Image __desc__: %s' % im)
    im.load()
    print('  Image __desc__: %s' % im)
    print('URL download: PASSED')

    # Valid URL but without an image extension 
    im = Image(url='http://bit.ly/great_horned_owl')
    print('  Image __desc__: %s' % im)
    im.load()
    print('  Image __desc__: %s' % im)


    # Invalid URL with ignore
    im = Image(url='https://a_bad_url.jpg')
    print('  Image __desc__: %s' % im)
    im.load(ignoreErrors=False)
    print('  Image __desc__: %s' % im)
    print('Invalid URL download: PASSED')

    # URL with filename 
    im = Image(url='https://upload.wikimedia.org/wikipedia/commons/thumb/2/23/Bubo_virginianus_06.jpg/1920px-Bubo_virginianus_06.jpg', filename='/tmp/myfile.jpg')
    print('  Image __desc__: %s' % im)
    print('  Image length: %d' %  len(im))
    im.download()
    print('  Image __desc__: %s' % im)
    im.load()
    print('  Image __desc__: %s' % im)
    print('URL with filename download: PASSED')

    # Filename object
    im = ImageDetection(filename='jebyrne.jpg', xmin=100, ymin=100, width=700, height=1000, category='face')
    print('Image __desc__: %s' % im)
    #im.show(figure=1)
    im.crop()
    print('Image __desc__: %s' % im)
    print('Filename: PASSED')

    # Array constructor
    

if __name__ == "__main__":
    run()






import numpy as np
import math
import numpy.linalg
import scipy.spatial
from itertools import product


def covariance_to_ellipse(cov):
    """2x2 covariance matrix to rotated bounding box"""
    (d,V) = numpy.linalg.eig(cov)
    return (d[0], d[1], math.atan2(V[1,0],V[0,0]))  # (major_axis_len, minor_axis_len, angle_in_radians)

def dehomogenize(p):
    return np.float32(p[0:-1, :]) / np.float32(p[-1,:])

def homogenize(p):
    return np.float32(np.vstack( (p, np.ones((1,p.shape[1]))) ))

def apply_homography(H,p):
    return dehomogenize(H*homogenize(p))

def similarity_imtransform2D(c=(0,0), r=0, s=1):
    # cv2 rotation is in degrees
    import cv2
    deg = r * 180. / math.pi
    A = cv2.getRotationMatrix2D(c, deg, s)
    return A

def similarity_imtransform(txy=(0,0), r=0, s=1):
    R = np.mat([[np.cos(r), -np.sin(r), 0], [np.sin(r), np.cos(r), 0], [0,0,1]])
    S = np.mat([[s,0,0], [0, s, 0], [0,0,1]])
    T = np.mat([[0,0,txy[0]], [0,0,txy[1]], [0,0,0]])
    return S*R + T  # composition


def affine_imtransform(txy=(0,0), r=0, sx=1, sy=1, kx=0, ky=0):
    R = np.mat([[np.cos(r), -np.sin(r), 0], [np.sin(r), np.cos(r), 0], [0,0,1]])
    S = np.mat([[sx,0,0], [0, sy, 0], [0,0,1]])
    K = np.mat([[1,ky,0], [kx,1,0], [0,0,1]])
    T = np.mat([[0,0,txy[0]], [0,0,txy[1]], [0,0,0]])
    return K*S*R + T  # composition

def random_affine_imtransform(txy=((0,0),(0,0)), r=(0,0), sx=(1,1), sy=(1,1), kx=(0,0), ky=(0,0)):
    return affine_imtransform(txy=(uniform_random_in_range(txy[0]), uniform_random_in_range(txy[1])),
                              r=uniform_random_in_range(r),
                              sx=uniform_random_in_range(sx),
                              sy=uniform_random_in_range(sy),
                              kx=uniform_random_in_range(kx),
                              ky=uniform_random_in_range(ky))

def imtransform2D(im, A):
    import cv2
    return cv2.warpAffine(im, A, (im.shape[1], im.shape[0]))

def imtransform(im, A):
    import cv2
    # cv2.warpPerspective(src, M, dsize[, dst[, flags[, borderMode[, borderValue]]]]) -> dst
    return cv2.warpPerspective(im, A, (im.shape[1], im.shape[0]))

def frame_to_bbox(fr):
    """ bbox = [ (xmin,ymin,width,height), ... ] """
    return [ (x[0]-x[2]/2, x[1]-x[2]/2, x[0]+x[2]/2, x[1]+x[2]/2) for x in fr]

def sub2ind(i, j, rows):
    return (i*rows + j)

def sqdist(d_obs, d_ref):
    """Given d_obs of size (MxD) ad d_ref of size (NxD) return (MxN) matrix of pairwise euclidean distances"""
    return scipy.spatial.distance.cdist(np.array(d_obs), np.array(d_ref), 'euclidean')  # small number of descriptors only

def normalize(x):
    return x / (np.linalg.norm(x.astype(np.float64))+1E-16)




class BoundingBox():
    def __init__(self, xmin=None, ymin=None, xmax=None, ymax=None, centroid=None, xcentroid=None, ycentroid=None, width=None, height=None, label=None, score=None, mask=None):

        if xmin is not None and ymin is not None and xmax is not None and ymax is not None:
            self.xmin = float(xmin)
            self.ymin = float(ymin)
            self.xmax = float(xmax)
            self.ymax = float(ymax)
        elif xmin is not None and ymin is not None and width is not None and height is not None:
            self.xmin = float(xmin)
            self.ymin = float(ymin)
            self.xmax = self.xmin + float(width)
            self.ymax = self.ymin + float(height)
        elif centroid is not None and width is not None and height is not None:
            self.xmin = float(centroid[0]) - float(width)/2.0
            self.ymin = float(centroid[1]) - float(height)/2.0
            self.xmax = float(centroid[0]) + float(width)/2.0
            self.ymax = float(centroid[1]) + float(height)/2.0                                                
        elif xcentroid is not None and ycentroid is not None and width is not None and height is not None:
            self.xmin = float(xcentroid) - float(width)/2.0
            self.ymin = float(ycentroid) - float(height)/2.0
            self.xmax = float(xcentroid) + float(width)/2.0
            self.ymax = float(ycentroid) + float(height)/2.0                                                
        elif mask is not None:
            # Convex hull of non-zero pixels in a mask image
            imx = np.sum(mask, axis=0)
            imy = np.sum(mask, axis=1)            
            self.xmin = np.argwhere(imx > 0)[0]
            self.ymin = np.argwhere(imy > 0)[0]
            self.xmax = np.argwhere(imx > 0)[-1]
            self.ymax = np.argwhere(imy > 0)[-1]
        else:
            self.xmin = float('nan')
            self.ymin = float('nan')
            self.xmax = float('nan')
            self.ymax = float('nan')

        self.label = label
        self.score = score

    def clone(self):
        return BoundingBox(xmin=self.xmin, xmax=self.xmax, ymin=self.ymin, ymax=self.ymax, label=self.label, score=self.score)
        
    def __eq__(self, other):
        """Bounding box equality"""
        return self.xmin==other.xmin and self.xmax==other.xmax and self.ymin==other.ymin and self.ymax==other.ymax

    def __neq__(self, other):
        """Bounding box non-equality"""
        return not self.__eq__(other)

    def __repr__(self):
        xmin = '%1.1f' % float(self.xmin) if not np.isnan(self.xmin) else 'nan'
        ymin = '%1.1f' % float(self.ymin) if not np.isnan(self.ymin) else 'nan'
        xmax = '%1.1f' % float(self.xmax) if not np.isnan(self.xmax) else 'nan'
        ymax = '%1.1f' % float(self.ymax) if not np.isnan(self.ymax) else 'nan'                
        return str('<strpy.boundingbox: xmin=%s, ymin=%s, xmax=%s, ymax=%s>'% (xmin, ymin, xmax, ymax))

    def __str__(self):
        return self.__repr__()

    def invalid(self):
        is_undefined = np.isnan(self.xmin) or np.isnan(self.ymin) or np.isnan(self.xmax) or np.isnan(self.ymax)
        is_degenerate =  ((self.xmax-self.xmin)<1) or ((self.ymax-self.ymin)<1) or (self.xmin >= self.xmax) or (self.ymin >= self.ymax)
        return is_undefined or is_degenerate

    def translate(self, dx=0, dy=0):
        self.xmin = self.xmin + dx
        self.ymin = self.ymin + dy
        self.xmax = self.xmax + dx
        self.ymax = self.ymax + dy
        return self

    def valid(self):
        return not self.invalid()

    def isvalid(self):
        return self.valid()
            
    def width(self):
        return self.xmax - self.xmin

    def setwidth(self, w):
        if w <= 0:
            raise ValueError('invalid width')
        self.xmax += (w - self.width()) / 2.0
        self.xmin -= (w - self.width()) / 2.0        
        return self
    
    def setheight(self, h):
        if h <= 0:
            raise ValueError('invalid height')
        self.ymax += (h - self.height()) / 2.0
        self.ymin -= (h - self.height()) / 2.0        
        return self
        
    def height(self):
        return self.ymax - self.ymin
    
    def centroid(self):
        """(x,y) tuple of centroid"""
        return [self.xmin + (float(self.width())/2.0), self.ymin + (float(self.height())/2.0)]
            
    def x_centroid(self):
        return self.centroid()[0]

    def y_centroid(self):
        return self.centroid()[1]
        
    def area(self):
        return self.width() * self.height()

    def to_xywh(self):
        """Convert corners to (x,y,width,height) format"""
        return [self.xmin, self.ymin, self.width(), self.height()]

    def dx(self, bb):
        return bb.xmin - self.xmin

    def dy(self, bb):
        return bb.ymin - self.ymin

    def sqdist(self, bb):
        """Squared distance between upper left corners of two bounding boxes"""
        return np.power(self.dx(bb), 2.0) + np.power(self.dy(bb), 2.0)

    def dist(self, bb):
        """Distance between centroids of two bounding boxes"""
        return np.sqrt(np.sum(np.square(np.array(bb.centroid()) - np.array(self.centroid()))))
    
    def islabel(self, label):
        return self.label.lower() == label.lower()
    
    def overlap(self, bb):
        """area of intersection / area of union"""
        """NOTE: this is poorly named and should be changed"""
        if bb.invalid() or self.invalid():
            return 0.0
        w = min(self.xmax, bb.xmax) - max(self.xmin, bb.xmin);
        h = min(self.ymax, bb.ymax) - max(self.ymin, bb.ymin);
        if ((w < 0) or (h < 0)):
            iou = 0;   # invalid (no overlap)
        else:
            area_intersection = w*h;
            area_union = (self.area() + bb.area() - area_intersection);
            iou = area_intersection / area_union;
        return iou;

    def iou(self, bb):
        """area of intersection / area of union"""
        if bb is None:
            return 0.0
        return self.overlap(bb)

    def area_of_intersection(self, bb):
        """area of intersection"""
        if bb.invalid() or self.invalid():
            return 0.0
        w = min(self.xmax, bb.xmax) - max(self.xmin, bb.xmin);
        h = min(self.ymax, bb.ymax) - max(self.ymin, bb.ymin);
        if ((w < 0) or (h < 0)):
            aoi = 0;   # invalid (no overlap)
        else:
            aoi = w*h;
        return aoi

    def cover(self, bb):
        """Fraction of this bounding box covered by other bbox"""
        return self.area_of_intersection(bb) / float(self.area())

    def intersection(self, bb):
        """Intersection of two bounding boxes"""
        self.xmin = max(bb.xmin, self.xmin)
        self.ymin = max(bb.ymin, self.ymin)
        self.xmax = min(bb.xmax, self.xmax)
        self.ymax = min(bb.ymax, self.ymax)
        return self
                    
    def union(self, bb):
        """Union of two bounding boxes"""
        self.xmin = min(bb.xmin, self.xmin)
        self.ymin = min(bb.ymin, self.ymin)
        self.xmax = max(bb.xmax, self.xmax)
        self.ymax = max(bb.ymax, self.ymax)
        return self
        
    def inside(self, p):
        return (p[0] >= self.xmin) and (p[1] >= self.ymin) and (p[0] <= self.xmax) and (p[1] <= self.ymax)

    def dilate(self, scale=1):
        """Change scale of bounding box keeping centroid constant"""
        w = self.width()
        h = self.height()
        c = self.centroid()
        old_x = self.xmin
        old_y = self.ymin
        new_x = (float(w)/2.0) * scale
        new_y = (float(h)/2.0) * scale
        self.xmin = c[0] - new_x
        self.ymin = c[1] - new_y
        self.xmax = c[0] + new_x
        self.ymax = c[1] + new_y
        # new_x and new_y are the offsets of the original bounding box in the new bounding box.
        # Can use this to use an extracted chip in place of a full image when we're cropping something
        # with dilation.
        return self

    def dilate_height(self, scale=1):
        """Change scale of bounding box in y direction keeping centroid constant"""
        h = self.height()
        c = self.centroid()
        self.ymin = c[1]-(float(h)/2.0)*scale
        self.ymax = c[1]+(float(h)/2.0)*scale
        return self

    def dilate_width(self, scale=1):
        """Change scale of bounding box in x direction keeping centroid constant"""
        w = self.width()
        c = self.centroid()
        self.xmin = c[0]-(float(w)/2.0)*scale
        self.xmax = c[0]+(float(w)/2.0)*scale
        return self

    def dilate_topheight(self, scale=1):
        """Change scale of bounding box in positive y direction only, changing centroid"""
        h = self.height()
        c = self.centroid()
        self.ymin = c[1]-(float(h)/2.0)*scale
        return self

    def rescale(self, scale=1):
        self.xmin = scale * self.xmin
        self.ymin = scale * self.ymin
        self.xmax = scale * self.xmax
        self.ymax = scale * self.ymax                        
        return self

    def imscale(self, im):
        w = (1.0 / float(im.width()))
        h = (1.0 / float(im.height()))
        self.xmin = w * self.xmin
        self.ymin = h * self.ymin
        self.xmax = w * self.xmax
        self.ymax = h * self.ymax                        
        return self
    
    def maxsquare(self):
        w = max(self.width(), self.height())
        h = max(self.height(), self.width())
        c = self.centroid()
        self.xmin = c[0]-(float(w)/2.0)
        self.ymin = c[1]-(float(h)/2.0)
        self.xmax = c[0]+(float(w)/2.0)
        self.ymax = c[1]+(float(h)/2.0)
        return self
        
    def imclip(self, img):
        """Clip bounding box to image rectangle"""
        self.intersection(BoundingBox(xmin=0, ymin=0, xmax=img.shape[1], ymax=img.shape[0]))
        return self
        
    def convexhull(self, fr):
        self.xmin = np.min(fr[:,0])
        self.ymin = np.min(fr[:,1])
        self.xmax = np.max(fr[:,0])
        self.ymax = np.max(fr[:,1])
        return self

    def aspectratio(self):
        return float(self.height()) / float(self.width())

    def shape(self):
        return (self.height(), self.width())

    def mindimension(self):
        return np.min(self.shape)

class Ellipse():
    def __init__(self, semi_major, semi_minor, xcenter, ycenter, phi):
        """Ellipse parameterization, for length of semimajor (half width of ellipse) and semiminor axis (half height), center point and angle phi in radians"""
        self.major = semi_major
        self.minor = semi_minor
        self.center_x = xcenter
        self.center_y = ycenter
        self.phi = phi

    def __repr__(self):
        return str('<strpy.geometry.ellipse: semimajor=%s, semiminor=%s, xcenter=%s, ycenter=%s, phi=%s (rad)>'% (self.major, self.minor, self.center_x, self.center_y, self.phi))


    def area(self):
        return math.pi * self.major * self.minor

    def center(self):
        return (int(self.center_x), int(self.center_y))

    def axes(self):
        return (int(self.major), int(self.minor))

    def angle(self):
        """in degrees"""
        return int(self.phi * 180 / math.pi)

    def rescale(self, scale):
        self.major *= scale
        self.minor *= scale
        self.center_x *= scale
        self.center_y *= scale
        return self

    def estimate_boundingbox(self):
        """ Estimate an equivalent bounding box based on scaling to a common area.
        Note, this does not factor in rotation.
        (c*l)*(c*w) = a_e  --> c = \sqrt(a_e / a_r) """

        bbox = BoundingBox(width=2*self.major, height=2*self.minor, xcentroid=self.center_x, ycentroid=self.center_y)
        a_r = bbox.area()
        c = (self.area() / a_r) ** 0.5
        bbox2 = bbox.clone().dilate(c)
        # print 'start area bbox: %f' % bbox.area()
        # print 'end   area bbox: %f' % bbox2.area()
        # print 'area ellipse   : %f' % self.area()
        return bbox2

    def inside(self, p, y=None):
        """Return true if a point p=(x,y) is inside the ellipse"""
        if (self.phi != 0):
            raise ValueError('FIXME: inside only supported for phi=0')
        (x,y) = p if y is None else (p,y)
        return ((np.square(x - self.center_x) / np.square(self.major)) + (np.square(y - self.center_y) / np.square(self.minor))) <= 1

    def mask(self):
        """Return a binary mask of size equal to the bounding box such that the pixels correspond to the interior of the ellipse"""
        (H,W) = (int(np.round(2*self.minor)), int(np.round(2*self.major)))
        img = np.zeros( (H,W) )
        for (y,x) in product(range(0,H), range(0,W)):
            img[y,x] = self.inside(x,y)
            print (x,y)
        return img

def bbox_to_ellipse(bb):
    (xcenter,ycenter) = bb.centroid()
    return Ellipse(bb.width()/2.0, bb.height()/2.0, xcenter,ycenter,0)



import flickrapi  # optional
import flickrapi.shorturl
import urllib
import os
import re
import random
import math

flickr_api_key = os.environ.get('FLICKR_API_KEY')
#flickr = flickrapi.FlickrAPI(flickrapikey, flickrapisecret)

def search(searchtag):
    flickr = flickrapi.FlickrAPI(flickr_api_key)
    photos = flickr.walk(text=searchtag,per_page='500')  
    return (photos, flickr)

def download(tag='owl'):
    (photos, flickr) = search(tag)
    for img in photos:        
        id = img.get('id')
        info = flickr.photos_getInfo(photo_id=id)    
        url = 'http://farm'+img.get('farm')+'.staticflickr.com/'+img.get('server')+'/'+img.get('id')+'_'+img.get('secret')+'_n.jpg'
        imfile = "/tmp/"+tag.replace(',','_')+"_"+id+".jpg"
        urllib.urlretrieve(url, imfile)
